__author__ = 'labx'

import os
import random
import sys
import copy

import numpy
import xraylib
import h5py

try:
    from PyQt5.QtCore import QSettings
    from PyQt5.QtWidgets import QWidget, QGridLayout, QLabel
    from PyQt5.QtGui import QFont, QPalette, QColor
except:
    pass

from matplotlib.patches import FancyArrowPatch, ArrowStyle
from scipy import optimize
from numpy import asarray

try:
    from oasys.widgets import gui
except:
    pass

from oasys.widgets import congruence
from oasys.util.oasys_util import get_sigma, get_fwhm, get_average

try:
    import matplotlib
    import matplotlib.pyplot as plt
    from matplotlib import cm
    from matplotlib import figure as matfig
    import pylab
except ImportError:
    print(sys.exc_info()[1])
    pass

# import Shadow.ShadowToolsPrivate as stp
from shadow4.beam.s4_beam import S4Beam

import scipy.constants as codata

class ShadowCongruence():
    @classmethod
    def check_empty_beam(cls, input_beam):
        if input_beam is None: return False
        elif not hasattr(input_beam, "rays"): return False
        elif len(input_beam.rays) == 0: return False
        else: return True

    @classmethod
    def check_empty_data(cls, input_data):
        if input_data is None:  return False
        else:                   return cls.check_empty_beam(input_data.beam)

    @classmethod
    def check_good_beam(cls, input_beam):
        return len(input_beam.rays[numpy.where(input_beam.rays[:, 9] == 1)]) > 0

    @classmethod
    def checkBraggFile(cls, file_name):
        file = open(file_name, "r")

        try:
            rows = file.readlines()


            if len(rows) < 10: raise Exception("Bragg file malformed, please check input")



            if "# Bragg version," in rows[0]:
                pass # version 2
            else:
                first_row = ShadowCongruence.__get_numbers(rows[0].strip())

                if not len(first_row) == 3: raise Exception("Bragg file malformed, please check input")

                second_row = ShadowCongruence.__get_numbers(rows[1].strip())
                if not len(second_row) == 3: raise Exception("Bragg file malformed, please check input")

                if not (rows[2].strip().startswith("(") and \
                                rows[3].strip().startswith("(") and \
                                rows[4].strip().startswith("(") and \
                                rows[5].strip().startswith("(")):
                    raise Exception("Bragg file malformed, please check input")

                seventh_row = ShadowCongruence.__get_numbers(rows[6].strip())
                if not len(seventh_row) == 3: raise Exception("Bragg file malformed, please check input")

                eighth_row = ShadowCongruence.__get_numbers(rows[7].strip())
                if not len(eighth_row) == 3: raise Exception("Bragg file malformed, please check input")

                nineth_row = ShadowCongruence.__get_numbers(rows[8].strip())
                if not len(nineth_row) == 1: raise Exception("Bragg file malformed, please check input")
        except Exception as e:
            file.close()

            raise e

    @classmethod
    def checkPreReflFile(cls, file_name):
        file = open(file_name, "r")

        try:
            rows = file.readlines()

            if len(rows) < 3: raise Exception("PreRefl file malformed, please check input")

            first_row = ShadowCongruence.__get_numbers(rows[0].strip())
            if not len(first_row) == 4: raise Exception("PreRefl file malformed, please check input")

            second_row = ShadowCongruence.__get_numbers(rows[1].strip())
            if not len(second_row) == 1: raise Exception("PreRefl file malformed, please check input")

            try:
                elements = int(second_row[0])
            except:
                raise Exception("PreRefl file malformed, please check input")

            if len(rows) != (elements*2) + 2: raise Exception("PreRefl file malformed, please check input")
        except Exception as e:
            file.close()

            raise e

    @classmethod
    def checkPreMLayerFile(cls, file_name):
        file = open(file_name, "r")

        try:
            rows = file.readlines()

            if len(rows) < 2: raise Exception("PreMLayer file malformed, please check input")

            first_row = ShadowCongruence.__get_numbers(rows[0].strip())
            if not len(first_row) == 1: raise Exception("PreMLayer file malformed, please check input")

            try:
                elements = int(first_row[0])
            except:
                raise Exception("PreRefl file malformed, please check input")

            second_row = ShadowCongruence.__get_numbers(rows[1].strip())
            if not len(second_row) == int(elements): raise Exception("PreMLayer file malformed, please check input")

            try:
                separator_row = ShadowCongruence.__get_numbers(rows[2 + elements*3].strip())
                if not len(separator_row) == 1: raise Exception("PreMLayer file malformed, please check input")
            except:
                raise Exception("PreRefl file malformed, please check input")

            next_row = ShadowCongruence.__get_numbers(rows[2 + elements*3 + 1].strip())
            if not len(next_row) == 4: raise Exception("PreMLayer file malformed, please check input")
        except Exception as e:
            file.close()

            raise e

    @classmethod
    def check2ColumnFormatFile(cls, file_name, specific_name):
        try:
            if file_name.startswith('/'):
                values = numpy.loadtxt(os.path.abspath(file_name))
            else:
                values = numpy.loadtxt(os.path.abspath(os.path.join(os.path.curdir, file_name)))
        except:
            raise Exception(specific_name + " file malformed (should be 2 or more columns of numbers, separated by spaces), please check input")

        if len(values) < 2: raise Exception(specific_name + " file malformed (should be 2 or more columns of numbers, separated by spaces), please check input")

    @classmethod
    def checkErrorProfileFile(cls, file_name):
        file = open(file_name, "r")

        try:
            rows = file.readlines()

            if len(rows) < 2: raise Exception("Surface Error file malformed, please check input")

            first_row = ShadowCongruence.__get_numbers(rows[0].strip())
            if not len(first_row) == 2: raise Exception("Surface Error file malformed, please check input")

            n_x = int(first_row[0])

            if n_x > 500:
                raise Exception("Malformed file: maximum allowed point in X direction is 500")

        except Exception as e:
            file.close()

            raise e

    @classmethod
    def __get_numbers(cls, string):
        values = string.strip().split(" ")

        numbers = []
        for value in values:
            if not value == "":
                try:
                    numbers.append(value)
                except:
                    pass

        return numbers

class ShadowStatisticData:
    intensity = 0.0
    total_number_of_rays = 0
    total_good_rays = 0
    total_lost_rays = 0

    def __init__(self, intensity = 0.0,
                 total_number_of_rays = 0,
                 total_good_rays = 0,
                 total_lost_rays = 0):
        self.intensity = intensity
        self.total_number_of_rays = total_number_of_rays
        self.total_good_rays = total_good_rays
        self.total_lost_rays = total_lost_rays

class ShadowHistoData(ShadowStatisticData):
    fwhm = 0.0
    x_fwhm_i = 0.0
    x_fwhm_f = 0.0
    y_fwhm = 0.0

    def __init__(self, intensity = 0.0,
                 total_number_of_rays = 0,
                 total_good_rays = 0,
                 total_lost_rays = 0,
                 fwhm = 0.0,
                 x_fwhm_i = 0.0,
                 x_fwhm_f = 0.0,
                 y_fwhm = 0.0):
        super().__init__(intensity, total_number_of_rays, total_good_rays, total_lost_rays)
        self.fwhm = fwhm
        self.x_fwhm_i = x_fwhm_i
        self.x_fwhm_f = x_fwhm_f
        self.y_fwhm = y_fwhm

class ShadowPlotData(ShadowStatisticData):
    fwhm_h = 0.0
    fwhm_v = 0.0

    def __init__(self, intensity = 0.0,
                 total_number_of_rays = 0,
                 total_good_rays = 0,
                 total_lost_rays = 0,
                 fwhm_h = 0.0,
                 fwhm_v = 0.0):
        super().__init__(intensity, total_number_of_rays, total_good_rays, total_lost_rays)
        self.fwhm_h = fwhm_h
        self.fwhm_v = fwhm_v

try:
    class ShadowPlot:

        _is_conversione_active = True

        #########################################################################################
        #
        # FOR TEMPORARY USE: FIX AN ERROR IN PYMCA.PLOT.IMAGEWIEW
        #
        #########################################################################################


        @classmethod
        def set_conversion_active(cls, is_active=True):
            ShadowPlot._is_conversione_active = is_active

        @classmethod
        def get_conversion_active(cls):
            return ShadowPlot._is_conversione_active

        """Sample code to add 2D dataset saving as text to ImageView."""


        #########################################################################################
        #
        # WIDGET FOR DETAILED PLOT
        #
        #########################################################################################

        class InfoBoxWidget(QWidget):
            intensity_field = ""
            flux_field = ""
            total_rays_field = ""
            total_good_rays_field = ""
            total_lost_rays_field = ""
            fwhm_h_field = ""
            fwhm_v_field = ""
            sigma_h_field = ""
            sigma_v_field = ""
            centroid_h_field = ""
            centroid_v_field = ""

            def __init__(self, x_scale_factor = 1.0, y_scale_factor = 1.0, is_2d=True):
                super(ShadowPlot.InfoBoxWidget, self).__init__()

                info_box_inner=gui.widgetBox(self, "Info")
                info_box_inner.setFixedHeight(518*y_scale_factor)
                info_box_inner.setFixedWidth(230*x_scale_factor)

                self.flux_box = gui.widgetBox(info_box_inner, "", addSpace=False, orientation="horizontal")
                self.flux     = gui.lineEdit(self.flux_box, self, "flux_field", "\u03a6 [ph/s/0.1%BW]", tooltip="Flux", labelWidth=115, valueType=str, orientation="horizontal")
                self.flux_box.setVisible(False)

                self.intensity = gui.lineEdit(info_box_inner, self, "intensity_field", "Intensity", tooltip="Intensity", labelWidth=115, valueType=str, orientation="horizontal")
                self.total_rays = gui.lineEdit(info_box_inner, self, "total_rays_field", "Total Rays", tooltip="Total Rays", labelWidth=115, valueType=str, orientation="horizontal")
                self.total_good_rays = gui.lineEdit(info_box_inner, self, "total_good_rays_field", "Total Good Rays", tooltip="Total Good Rays", labelWidth=115, valueType=str, orientation="horizontal")
                self.total_lost_rays = gui.lineEdit(info_box_inner, self, "total_lost_rays_field", "Total Lost Rays", tooltip="Total Lost Rays", labelWidth=115, valueType=str, orientation="horizontal")

                label_box_1 = gui.widgetBox(info_box_inner, "", addSpace=False, orientation="horizontal")

                self.label_h = QLabel("FWHM ")
                self.label_h.setFixedWidth(115)
                palette =  QPalette(self.label_h.palette())
                palette.setColor(QPalette.Foreground, QColor('blue'))
                self.label_h.setPalette(palette)
                label_box_1.layout().addWidget(self.label_h)
                self.fwhm_h = gui.lineEdit(label_box_1, self, "fwhm_h_field", "", tooltip="FWHM", labelWidth=115, valueType=str, orientation="horizontal")

                if is_2d:
                    label_box_2 = gui.widgetBox(info_box_inner, "", addSpace=False, orientation="horizontal")

                    self.label_v = QLabel("FWHM ")
                    self.label_v.setFixedWidth(115)
                    palette =  QPalette(self.label_v.palette())
                    palette.setColor(QPalette.Foreground, QColor('red'))
                    self.label_v.setPalette(palette)
                    label_box_2.layout().addWidget(self.label_v)
                    self.fwhm_v = gui.lineEdit(label_box_2, self, "fwhm_v_field", "", tooltip="FWHM", labelWidth=115, valueType=str, orientation="horizontal")

                label_box_1 = gui.widgetBox(info_box_inner, "", addSpace=False, orientation="horizontal")

                self.label_s_h = QLabel("\u03c3 (s.d.)")
                self.label_s_h.setFixedWidth(115)
                palette =  QPalette(self.label_s_h.palette())
                palette.setColor(QPalette.Foreground, QColor('blue'))
                self.label_s_h.setPalette(palette)
                label_box_1.layout().addWidget(self.label_s_h)
                self.sigma_h = gui.lineEdit(label_box_1, self, "sigma_h_field", "", tooltip="Sigma", labelWidth=115, valueType=str, orientation="horizontal")

                if is_2d:
                    label_box_2 = gui.widgetBox(info_box_inner, "", addSpace=False, orientation="horizontal")

                    self.label_s_v = QLabel("\u03c3 (s.d.)")
                    self.label_s_v.setFixedWidth(115)
                    palette =  QPalette(self.label_s_v.palette())
                    palette.setColor(QPalette.Foreground, QColor('red'))
                    self.label_s_v.setPalette(palette)
                    label_box_2.layout().addWidget(self.label_s_v)
                    self.sigma_v = gui.lineEdit(label_box_2, self, "sigma_v_field", "", tooltip="Sigma", labelWidth=115, valueType=str, orientation="horizontal")

                label_box_1 = gui.widgetBox(info_box_inner, "", addSpace=False, orientation="horizontal")

                self.label_c_h = QLabel("centroid ")
                self.label_c_h.setFixedWidth(115)
                palette =  QPalette(self.label_c_h.palette())
                palette.setColor(QPalette.Foreground, QColor('blue'))
                self.label_c_h.setPalette(palette)
                label_box_1.layout().addWidget(self.label_c_h)
                self.centroid_h = gui.lineEdit(label_box_1, self, "centroid_h_field", "", tooltip="Centroid", labelWidth=115, valueType=str, orientation="horizontal")

                if is_2d:
                    label_box_2 = gui.widgetBox(info_box_inner, "", addSpace=False, orientation="horizontal")

                    self.label_c_v = QLabel("centroid ")
                    self.label_c_v.setFixedWidth(115)
                    palette =  QPalette(self.label_c_v.palette())
                    palette.setColor(QPalette.Foreground, QColor('red'))
                    self.label_c_v.setPalette(palette)
                    label_box_2.layout().addWidget(self.label_c_v)
                    self.centroid_v = gui.lineEdit(label_box_2, self, "centroid_v_field", "", tooltip="Sigma", labelWidth=115, valueType=str, orientation="horizontal")

                self.intensity.setReadOnly(True)
                font = QFont(self.intensity.font())
                font.setBold(True)
                self.intensity.setFont(font)
                palette = QPalette(self.intensity.palette())
                palette.setColor(QPalette.Text, QColor('dark blue'))
                palette.setColor(QPalette.Base, QColor(243, 240, 160))
                self.intensity.setPalette(palette)

                self.flux.setReadOnly(True)
                font = QFont(self.flux.font())
                font.setBold(True)
                self.flux.setFont(font)
                palette = QPalette(self.flux.palette())
                palette.setColor(QPalette.Text, QColor('dark blue'))
                palette.setColor(QPalette.Base, QColor(243, 240, 160))
                self.flux.setPalette(palette)

                self.total_rays.setReadOnly(True)
                font = QFont(self.total_rays.font())
                font.setBold(True)
                self.total_rays.setFont(font)
                palette = QPalette(self.intensity.palette())
                palette.setColor(QPalette.Text, QColor('dark blue'))
                palette.setColor(QPalette.Base, QColor(243, 240, 160))
                self.total_rays.setPalette(palette)

                self.total_good_rays.setReadOnly(True)
                font = QFont(self.total_good_rays.font())
                font.setBold(True)
                self.total_good_rays.setFont(font)
                palette = QPalette(self.total_good_rays.palette())
                palette.setColor(QPalette.Text, QColor('dark blue'))
                palette.setColor(QPalette.Base, QColor(243, 240, 160))
                self.total_good_rays.setPalette(palette)

                self.total_lost_rays.setReadOnly(True)
                font = QFont(self.total_lost_rays.font())
                font.setBold(True)
                self.total_lost_rays.setFont(font)
                palette = QPalette(self.total_lost_rays.palette())
                palette.setColor(QPalette.Text, QColor('dark blue'))
                palette.setColor(QPalette.Base, QColor(243, 240, 160))
                self.total_lost_rays.setPalette(palette)

                self.fwhm_h.setReadOnly(True)
                font = QFont(self.fwhm_h.font())
                font.setBold(True)
                self.fwhm_h.setFont(font)
                palette = QPalette(self.fwhm_h.palette())
                palette.setColor(QPalette.Text, QColor('dark blue'))
                palette.setColor(QPalette.Base, QColor(243, 240, 160))
                self.fwhm_h.setPalette(palette)

                self.sigma_h.setReadOnly(True)
                font = QFont(self.sigma_h.font())
                font.setBold(True)
                self.sigma_h.setFont(font)
                palette = QPalette(self.sigma_h.palette())
                palette.setColor(QPalette.Text, QColor('dark blue'))
                palette.setColor(QPalette.Base, QColor(243, 240, 160))
                self.sigma_h.setPalette(palette)

                self.centroid_h.setReadOnly(True)
                font = QFont(self.centroid_h.font())
                font.setBold(True)
                self.centroid_h.setFont(font)
                palette = QPalette(self.centroid_h.palette())
                palette.setColor(QPalette.Text, QColor('dark blue'))
                palette.setColor(QPalette.Base, QColor(243, 240, 160))
                self.centroid_h.setPalette(palette)

                if is_2d:
                    self.fwhm_v.setReadOnly(True)
                    font = QFont(self.fwhm_v.font())
                    font.setBold(True)
                    self.fwhm_v.setFont(font)
                    palette = QPalette(self.fwhm_v.palette())
                    palette.setColor(QPalette.Text, QColor('dark blue'))
                    palette.setColor(QPalette.Base, QColor(243, 240, 160))
                    self.fwhm_v.setPalette(palette)

                    self.sigma_v.setReadOnly(True)
                    font = QFont(self.sigma_v.font())
                    font.setBold(True)
                    self.sigma_v.setFont(font)
                    palette = QPalette(self.sigma_v.palette())
                    palette.setColor(QPalette.Text, QColor('dark blue'))
                    palette.setColor(QPalette.Base, QColor(243, 240, 160))
                    self.sigma_v.setPalette(palette)

                    self.centroid_v.setReadOnly(True)
                    font = QFont(self.centroid_v.font())
                    font.setBold(True)
                    self.centroid_v.setFont(font)
                    palette = QPalette(self.centroid_v.palette())
                    palette.setColor(QPalette.Text, QColor('dark blue'))
                    palette.setColor(QPalette.Base, QColor(243, 240, 160))
                    self.centroid_v.setPalette(palette)

            def set_flux(self, flux=None):
                if flux is None:
                    self.flux.setText("0.0")
                    self.flux_box.setVisible(False)
                else:
                    self.flux.setText('%.3E' % flux)
                    self.flux_box.setVisible(True)

            def clear(self):
                self.intensity.setText("0.0")
                self.flux.setText("0.0")
                self.flux_box.setVisible(False)
                self.total_rays.setText("0")
                self.total_good_rays.setText("0")
                self.total_lost_rays.setText("0")
                self.fwhm_h.setText("0.0000")
                if hasattr(self, "fwhm_v"):  self.fwhm_v.setText("0.0000")
                self.sigma_h.setText("0.0000")
                if hasattr(self, "sigma_v"):  self.sigma_v.setText("0.0000")
                self.centroid_h.setText("0.0000")
                if hasattr(self, "centroid_v"):  self.centroid_v.setText("0.0000")
                self.boundary_h.setText("")
                if hasattr(self, "boundary_v"):  self.centroid_v.setText("")

        class DetailedHistoWidget(QWidget):

            def __init__(self, x_scale_factor = 1.0, y_scale_factor = 1.0):
                super(ShadowPlot.DetailedHistoWidget, self).__init__()

                self.plot_canvas = gui.plotWindow(roi=False, control=False, position=True, logScale=True, fit=True)
                self.plot_canvas.setDefaultPlotLines(True)
                self.plot_canvas.setActiveCurveColor(color='blue')
                self.plot_canvas.setMinimumWidth(590*x_scale_factor)
                self.plot_canvas.setMaximumWidth(590*x_scale_factor)

                self.info_box = ShadowPlot.InfoBoxWidget(x_scale_factor, y_scale_factor, is_2d=False)

                layout = QGridLayout()

                layout.addWidget(   self.info_box, 0, 1, 1, 1)
                layout.addWidget(self.plot_canvas, 0, 0, 1, 1)

                layout.setColumnMinimumWidth(0, 600*x_scale_factor)
                layout.setColumnMinimumWidth(1, 230*x_scale_factor)

                self.setLayout(layout)

            def plot_histo(self, beam, col, nolost, xrange, ref, title, xtitle, ytitle, nbins = 100, xum="", ticket_to_add=None, flux=None):
                ticket = beam.histo1(col, nbins=nbins, xrange=xrange, nolost=nolost, ref=ref)
                if ref in [24, 25]: ticket['intensity'] = beam.get_column(ref, nolost=nolost).sum()

                # TODO: check congruence between tickets
                if not ticket_to_add is None:
                    last_ticket = copy.deepcopy(ticket)

                    if ticket['histogram'].size == ticket_to_add['histogram'].size:
                        ticket['histogram'] += ticket_to_add['histogram']
                        ticket['histogram_path'] += ticket_to_add['histogram_path']

                        ticket['intensity'] += ticket_to_add['intensity']
                        ticket['nrays'] += ticket_to_add['nrays']
                        ticket['good_rays'] += ticket_to_add['good_rays']

                ticket['fwhm'], ticket['fwhm_quote'], ticket['fwhm_coordinates'] = get_fwhm(ticket['histogram'], ticket['bin_center'], ret0=None)
                ticket['sigma']    = get_sigma(ticket['histogram'], ticket['bin_center'], ret0=numpy.nan)
                ticket['centroid'] = get_average(ticket['histogram'], ticket['bin_center'], ret0=numpy.nan)

                factor = ShadowPlot.get_factor(col)

                if ref != 0 and not ytitle is None:  ytitle = ytitle + ' weighted by ' + ShadowPlot.get_shadow_label(ref)

                histogram = ticket['histogram_path']
                bins      = ticket['bin_path']*factor

                if histogram.size > 1: self.plot_canvas.addCurve(bins, histogram, title, symbol='', color='blue', replace=True) #'+', '^', ','
                else:                  self.plot_canvas.clear()

                if not xtitle is None: self.plot_canvas.setGraphXLabel(xtitle)
                if not ytitle is None: self.plot_canvas.setGraphYLabel(ytitle)
                if not title is None: self.plot_canvas.setGraphTitle(title)
                self.plot_canvas.setInteractiveMode(mode='zoom')

                if ticket['fwhm'] is None: ticket['fwhm'] = 0.0
                if not ticket_to_add is None:
                    if last_ticket['fwhm'] is None: last_ticket['fwhm'] = 0.0

                n_patches = len(self.plot_canvas._backend.ax.patches)
                if (n_patches > 0): self.plot_canvas._backend.ax.patches.remove(self.plot_canvas._backend.ax.patches[n_patches-1])

                if not ticket['fwhm'] == 0.0:
                    x_fwhm_i, x_fwhm_f = ticket['fwhm_coordinates']
                    x_fwhm_i, x_fwhm_f = x_fwhm_i*factor, x_fwhm_f*factor
                    y_fwhm             = ticket['fwhm_quote']

                    self.plot_canvas._backend.ax.add_patch(FancyArrowPatch([x_fwhm_i, y_fwhm],
                                                              [x_fwhm_f, y_fwhm],
                                                              arrowstyle=ArrowStyle.CurveAB(head_width=2, head_length=4),
                                                              color='b',
                                                              linewidth=1.5))

                if histogram.size > 1:
                    if min(histogram) < 0: self.plot_canvas.setGraphYLimits(min(histogram), max(histogram))
                    else: self.plot_canvas.setGraphYLimits(0, max(histogram))

                self.plot_canvas.replot()

                self.info_box.intensity.setText("{:4.3f}".format(ticket['intensity']))
                self.info_box.set_flux(flux)
                self.info_box.total_rays.setText(str(ticket['nrays']))
                self.info_box.total_good_rays.setText(str(ticket['good_rays']))
                self.info_box.total_lost_rays.setText(str(ticket['nrays']-ticket['good_rays']))
                self.info_box.fwhm_h.setText("{:5.4f}".format(ticket['fwhm']*factor))
                self.info_box.label_h.setText("FWHM " + xum)
                self.info_box.sigma_h.setText("{:5.4f}".format(ticket['sigma']*factor))
                self.info_box.label_s_h.setText("\u03c3 " + xum)
                self.info_box.centroid_h.setText("{:5.4f}".format(ticket['centroid']*factor))
                self.info_box.label_c_h.setText("centroid " + xum)

                if not ticket_to_add is None:
                    return ticket, last_ticket
                else:
                    return ticket, None

            def clear(self):
                self.plot_canvas.clear()
                self.info_box.clear()

        class DetailedPlotWidget(QWidget):
            def __init__(self, x_scale_factor = 1.0, y_scale_factor = 1.0):
                super(ShadowPlot.DetailedPlotWidget, self).__init__()

                self.x_scale_factor = x_scale_factor
                self.y_scale_factor = y_scale_factor

                self.plot_canvas = gui.imageWiew(parent=self)

                self.plot_canvas.setColormap({"name":"temperature", "normalization":"linear", "autoscale":True, "vmin":0, "vmax":0, "colors":256})
                self.plot_canvas.setMinimumWidth(590 * x_scale_factor)
                self.plot_canvas.setMaximumWidth(590 * y_scale_factor)

                self.info_box = ShadowPlot.InfoBoxWidget(x_scale_factor, y_scale_factor)

                layout = QGridLayout()

                layout.addWidget(self.info_box,    0, 1, 1, 1)
                layout.addWidget(self.plot_canvas, 0, 0, 1, 1)

                layout.setColumnMinimumWidth(0, 600*x_scale_factor)
                layout.setColumnMinimumWidth(1, 230*x_scale_factor)

                self.setLayout(layout)

            def plot_xy(self, beam, var_x, var_y, title, xtitle, ytitle, xrange=None, yrange=None, nolost=1, nbins=100, nbins_h=None, nbins_v=None, xum="", yum="", ref=23, is_footprint=False, ticket_to_add=None, flux=None):
                matplotlib.rcParams['axes.formatter.useoffset']='False'

                if nbins_h is None: nbins_h = nbins
                if nbins_v is None: nbins_v = nbins

                ticket = beam.histo2(var_x, var_y, nbins=nbins, nbins_h=nbins_h, nbins_v=nbins_v, xrange=xrange, yrange=yrange, nolost=nolost, ref=ref)
                if ref in [24, 25]: ticket['intensity'] = beam.get_column(ref, nolost=nolost).sum()

                # TODO: check congruence between tickets
                if not ticket_to_add is None:
                    last_ticket = copy.deepcopy(ticket)

                    if ticket['histogram'].shape == ticket_to_add['histogram'].shape:
                        ticket['histogram'] += ticket_to_add['histogram']
                        ticket['histogram_h'] += ticket_to_add['histogram_h']
                        ticket['histogram_v'] += ticket_to_add['histogram_v']

                        ticket['intensity'] += ticket_to_add['intensity']
                        ticket['nrays'] += ticket_to_add['nrays']
                        ticket['good_rays'] += ticket_to_add['good_rays']

                ticket['fwhm_h'], ticket['fwhm_quote_h'], ticket['fwhm_coordinates_h'] = get_fwhm(ticket['histogram_h'], ticket['bin_h_center'], ret0=None)
                ticket['fwhm_v'], ticket['fwhm_quote_v'], ticket['fwhm_coordinates_v'] = get_fwhm(ticket['histogram_v'], ticket['bin_v_center'], ret0=None)
                ticket['sigma_h']    = get_sigma(ticket['histogram_h'], ticket['bin_h_center'], ret0=numpy.nan)
                ticket['sigma_v']    = get_sigma(ticket['histogram_v'], ticket['bin_v_center'], ret0=numpy.nan)
                ticket['centroid_h'] = get_average(ticket['histogram_h'], ticket['bin_h_center'], ret0=numpy.nan)
                ticket['centroid_v'] = get_average(ticket['histogram_v'], ticket['bin_v_center'], ret0=numpy.nan)

                factor1 = 1.0 if is_footprint else ShadowPlot.get_factor(var_x)
                factor2 = 1.0 if is_footprint else ShadowPlot.get_factor(var_y)

                xx = ticket['bin_h_edges']
                yy = ticket['bin_v_edges']

                if ticket['histogram'].size > 1:
                    xmin, xmax = xx.min(), xx.max()
                    ymin, ymax = yy.min(), yy.max()

                    origin = (xmin*factor1, ymin*factor2)
                    scale = (abs((xmax-xmin)/nbins_h)*factor1, abs((ymax-ymin)/nbins_v)*factor2)

                    self.plot_canvas.setColormap({"name":QSettings().value("output/shadow-default-colormap", "temperature", str),
                                                  "normalization":"linear",
                                                  "autoscale":True,
                                                  "vmin":0,
                                                  "vmax":0,
                                                  "colors":256})

                    # PyMCA inverts axis!!!! histogram must be calculated reversed
                    self.plot_canvas.setImage(ticket['histogram'].T, origin=origin, scale=scale)
                else:
                    self.plot_canvas.clear()

                if xtitle is None: xtitle=ShadowPlot.get_shadow_label(var_x)
                if ytitle is None: ytitle=ShadowPlot.get_shadow_label(var_y)

                self.plot_canvas.setGraphXLabel(xtitle)
                self.plot_canvas.setGraphYLabel(ytitle)
                self.plot_canvas.setGraphTitle(title)

                self.plot_canvas._histoHPlot.setGraphYLabel('A.U.')

                self.plot_canvas._histoHPlot._backend.ax.xaxis.get_label().set_color('white')
                self.plot_canvas._histoHPlot._backend.ax.xaxis.get_label().set_fontsize(1)
                for label in self.plot_canvas._histoHPlot._backend.ax.xaxis.get_ticklabels():
                    label.set_color('white')
                    label.set_fontsize(1)

                self.plot_canvas._histoVPlot.setGraphXLabel('A.U.')

                self.plot_canvas._histoVPlot._backend.ax.yaxis.get_label().set_color('white')
                self.plot_canvas._histoVPlot._backend.ax.yaxis.get_label().set_fontsize(1)
                for label in self.plot_canvas._histoVPlot._backend.ax.yaxis.get_ticklabels():
                    label.set_color('white')
                    label.set_fontsize(1)

                if ticket['fwhm_h'] is None: ticket['fwhm_h'] = 0.0
                if ticket['fwhm_v'] is None: ticket['fwhm_v'] = 0.0
                if not ticket_to_add is None:
                    if last_ticket['fwhm_h'] is None: last_ticket['fwhm_h'] = 0.0
                    if last_ticket['fwhm_v'] is None: last_ticket['fwhm_v'] = 0.0

                n_patches = len(self.plot_canvas._histoHPlot._backend.ax.patches)
                if (n_patches > 0): self.plot_canvas._histoHPlot._backend.ax.patches.remove(self.plot_canvas._histoHPlot._backend.ax.patches[n_patches-1])

                if not ticket['fwhm_h'] == 0.0:
                    x_fwhm_i, x_fwhm_f = ticket['fwhm_coordinates_h']
                    x_fwhm_i, x_fwhm_f = x_fwhm_i*factor1, x_fwhm_f*factor1
                    y_fwhm = ticket['fwhm_quote_h']

                    self.plot_canvas._histoHPlot._backend.ax.add_patch(FancyArrowPatch([x_fwhm_i, y_fwhm],
                                                                         [x_fwhm_f, y_fwhm],
                                                                         arrowstyle=ArrowStyle.CurveAB(head_width=2, head_length=4),
                                                                         color='b',
                                                                         linewidth=1.5))

                n_patches = len(self.plot_canvas._histoVPlot._backend.ax.patches)
                if (n_patches > 0): self.plot_canvas._histoVPlot._backend.ax.patches.remove(self.plot_canvas._histoVPlot._backend.ax.patches[n_patches-1])

                if not ticket['fwhm_v'] == 0.0:
                    y_fwhm_i, y_fwhm_f = ticket['fwhm_coordinates_v']
                    y_fwhm_i, y_fwhm_f = y_fwhm_i*factor2, y_fwhm_f*factor2
                    x_fwhm = ticket['fwhm_quote_v']

                    self.plot_canvas._histoVPlot._backend.ax.add_patch(FancyArrowPatch([x_fwhm, y_fwhm_i],
                                                                         [x_fwhm, y_fwhm_f],
                                                                         arrowstyle=ArrowStyle.CurveAB(head_width=2, head_length=4),
                                                                         color='r',
                                                                         linewidth=1.5))

                self.plot_canvas._histoHPlot.replot()
                self.plot_canvas._histoVPlot.replot()
                self.plot_canvas.replot()

                self.info_box.intensity.setText("{:4.3f}".format(ticket['intensity']))
                self.info_box.set_flux(flux)
                self.info_box.total_rays.setText(str(ticket['nrays']))
                self.info_box.total_good_rays.setText(str(ticket['good_rays']))
                self.info_box.total_lost_rays.setText(str(ticket['nrays']-ticket['good_rays']))
                self.info_box.fwhm_h.setText("{:5.4f}".format(ticket['fwhm_h'] * factor1))
                self.info_box.fwhm_v.setText("{:5.4f}".format(ticket['fwhm_v'] * factor2))
                self.info_box.label_h.setText("FWHM " + xum)
                self.info_box.label_v.setText("FWHM " + yum)
                self.info_box.sigma_h.setText("{:5.4f}".format(ticket['sigma_h'] * factor1))
                self.info_box.sigma_v.setText("{:5.4f}".format(ticket['sigma_v'] * factor2))
                self.info_box.label_s_h.setText("\u03c3 " + xum)
                self.info_box.label_s_v.setText("\u03c3 " + yum)
                self.info_box.centroid_h.setText("{:5.4f}".format(ticket['centroid_h'] * factor1))
                self.info_box.centroid_v.setText("{:5.4f}".format(ticket['centroid_v'] * factor2))
                self.info_box.label_c_h.setText("centroid " + xum)
                self.info_box.label_c_v.setText("centroid " + yum)

                if not ticket_to_add is None:
                    return ticket, last_ticket
                else:
                    return ticket, None

            def clear(self):
                self.plot_canvas.clear()

                self.plot_canvas._histoHPlot.clear()
                self.plot_canvas._histoVPlot.clear()

                self.plot_canvas._histoHPlot._backend.ax.xaxis.get_label().set_color('white')
                self.plot_canvas._histoHPlot._backend.ax.xaxis.get_label().set_fontsize(1)
                for label in self.plot_canvas._histoHPlot._backend.ax.xaxis.get_ticklabels():
                    label.set_color('white')
                    label.set_fontsize(1)

                self.plot_canvas._histoVPlot._backend.ax.yaxis.get_label().set_color('white')
                self.plot_canvas._histoVPlot._backend.ax.yaxis.get_label().set_fontsize(1)
                for label in self.plot_canvas._histoVPlot._backend.ax.yaxis.get_ticklabels():
                    label.set_color('white')
                    label.set_fontsize(1)

                self.plot_canvas._histoHPlot.setGraphYLabel('A.U.')
                self.plot_canvas._histoVPlot.setGraphXLabel('A.U.')

                self.plot_canvas._histoHPlot.replot()
                self.plot_canvas._histoVPlot.replot()

                self.info_box.clear()

        #########################################################################################

        @classmethod
        def plotxy_preview(cls, plot_window, beam, var_x, var_y, nolost=0, title='PLOTXY', xtitle=None, ytitle=None, is_footprint=False):

            matplotlib.rcParams['axes.formatter.useoffset']='False'

            col1 = beam.get_column(var_x, nolost=nolost)
            col2 = beam.get_column(var_y, nolost=nolost)

            factor1 = 1.0 if is_footprint else ShadowPlot.get_factor(var_x)
            factor2 = 1.0 if is_footprint else ShadowPlot.get_factor(var_y)

            if xtitle is None: xtitle=ShadowPlot.get_shadow_label(var_x)
            if ytitle is None: ytitle=ShadowPlot.get_shadow_label(var_y)

            plot_window.addCurve(col1*factor1, col2*factor2, title, symbol='.', color='blue', replace=True) #'+', '^', ','

            if not xtitle is None: plot_window.setGraphXLabel(xtitle)
            if not ytitle is None: plot_window.setGraphYLabel(ytitle)
            if not title is None: plot_window.setGraphTitle(title)
            plot_window.setInteractiveMode(mode='zoom')

        @classmethod
        def plot_histo_preview(cls, plot_window, beam, col, nolost, ref, title, xtitle, ytitle, conv=1.0):

            matplotlib.rcParams['axes.formatter.useoffset']='False'

            factor = ShadowPlot.get_factor(col)

            ticket = beam.histo1(col, nbins=100, xrange=None, nolost=nolost, ref=ref)

            if ref != 0 and not ytitle is None:  ytitle = ytitle + ' weighted by ' + ShadowPlot.get_shadow_label(ref)

            histogram = ticket['histogram_path']
            bins = ticket['bin_path']*factor

            plot_window.addCurve(bins, histogram, title, symbol='', color='blue', replace=True) #'+', '^', ','
            if not xtitle is None: plot_window.setGraphXLabel(xtitle)
            if not ytitle is None: plot_window.setGraphYLabel(ytitle)
            if not title is None: plot_window.setGraphTitle(title)
            plot_window.setInteractiveMode(mode='zoom')
            if min(histogram) < 0:
                plot_window.setGraphYLimits(min(histogram), max(histogram))
            else:
                plot_window.setGraphYLimits(0, max(histogram))
            plot_window.replot()

        @classmethod
        def get_factor(cls, var):
            if ShadowPlot.get_conversion_active() and var in [1, 2, 3, 4, 5, 6]: return 1e6 # m to micron
            else:                                                                return 1.0

        @classmethod
        def get_shadow_label(cls, var):
            return S4Beam.column_names_formatted()[var-1]
            # return (stp.getLabel(var-1))[0]


        #########################################################################################
        #
        # SAVE/LOAD FILES in HDF5 Format
        #
        #########################################################################################


        class PlotXYHdf5File(h5py.File):
            def __init__(self, file_name, mode="w"):
                try:
                    super(ShadowPlot.PlotXYHdf5File, self).__init__(name=file_name, mode=mode)
                except OSError as e:
                    if "already open" in str(e) and mode=="w":
                        super(ShadowPlot.PlotXYHdf5File, self).__init__(name=file_name, mode="a")
                        self.close()
                        super(ShadowPlot.PlotXYHdf5File, self).__init__(name=file_name, mode="w")

                if mode != "r":
                    self.coordinates = self.create_group("coordinates")
                    self.plots = self.create_group("xy_plots")
                    self.additional_data = self.create_group("additional_data")
                    self.last_plot = self.plots.create_group("last_plot")

                    self.has_last_plot = False
                    self.has_coordinate = False

                    self.attrs["default"]          = "coordinates/X"
                    self.attrs["file_name"]        = file_name
                    self.attrs["file_time"]        = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
                    self.attrs["creator"]          = "PlotXYHdf5File.__init__"
                    self.attrs["code"]             = "ShadowOui"
                    self.attrs["HDF5_Version"]     = h5py.version.hdf5_version
                    self.attrs["h5py_version"]     = h5py.version.version

            def get_last_plot(self, dataset_name="intensity"):
                return self.get_plot_xy(dataset_name=dataset_name)

            def get_coordinates(self):
                bin_h_center = self["coordinates/X"][()]
                h_label = self["coordinates"].attrs["x_label"]
                bin_v_center = self["coordinates/Y"][()]
                v_label = self["coordinates"].attrs["y_label"]

                return bin_h_center, bin_v_center, h_label, v_label

            def get_plot_xy(self, plot_name="last_plot", dataset_name="intensity"):
                histogram = self["/xy_plots/" + plot_name + "/" + dataset_name][()]
                histogram_h = self["/xy_plots/" + plot_name + "/histogram_h"][()]
                histogram_v = self["/xy_plots/" + plot_name + "/histogram_v"][()]

                return histogram, histogram_h, histogram_v, self["/xy_plots/" + plot_name].attrs

            def write_coordinates(self, ticket):
                if not self.has_coordinate:
                    self.x = self.coordinates.create_dataset("X", data=ticket["bin_h_center"])
                    self.y = self.coordinates.create_dataset("Y", data=ticket["bin_v_center"])
                    self.has_coordinate = True
                else:
                    self.x[...] = ticket["bin_h_center"]
                    self.y[...] = ticket["bin_v_center"]

                try:
                    self.coordinates.attrs["x_label"] = ShadowPlot.get_shadow_label(ticket["col_h"])
                    self.coordinates.attrs["y_label"] = ShadowPlot.get_shadow_label(ticket["col_v"])
                except:
                    self.coordinates.attrs["x_label"] = ticket["h_label"]
                    self.coordinates.attrs["y_label"] = ticket["v_label"]


            def add_plot_xy(self, ticket, plot_name="last_plot", dataset_name="intensity", attributes={}):
                    if plot_name is None or plot_name.strip() == "" or plot_name.strip() == "last_plot":
                        if not self.has_last_plot:
                            self.lp_histogram   = self.last_plot.create_dataset(dataset_name, data=ticket['histogram'])
                            self.lp_histogram_h = self.last_plot.create_dataset("histogram_h", data=ticket['histogram_h'])
                            self.lp_histogram_v = self.last_plot.create_dataset("histogram_v", data=ticket['histogram_v'])
                            self.has_last_plot = True
                        else:
                            if self.lp_histogram.name != "/xy_plots/last_plot/" + dataset_name:
                                self.last_plot.move(self.lp_histogram.name, "/xy_plots/last_plot/" + dataset_name)

                            self.lp_histogram[...]   = ticket['histogram']
                            self.lp_histogram_h[...] = ticket['histogram_h']
                            self.lp_histogram_v[...] = ticket['histogram_v']

                        self.last_plot.attrs["intensity"] = ticket["intensity"]
                        self.last_plot.attrs["total_rays"] = ticket["nrays"]
                        self.last_plot.attrs["good_rays"] = ticket["good_rays"]
                        self.last_plot.attrs["lost_rays"] = ticket["nrays"]-ticket["good_rays"]

                        if not attributes is None:
                            for key in attributes.keys():
                                self.last_plot.attrs[key] = attributes[key]

                    else:
                        plot = self.plots.create_group(plot_name)

                        plot.create_dataset(dataset_name, data=ticket['histogram'])
                        plot.create_dataset("histogram_h", data=ticket['histogram_h'])
                        plot.create_dataset("histogram_v", data=ticket['histogram_v'])
                        plot.attrs["intensity"] = ticket["intensity"]
                        plot.attrs["total_rays"] = ticket["nrays"]
                        plot.attrs["good_rays"] = ticket["good_rays"]
                        plot.attrs["lost_rays"] = ticket["nrays"]-ticket["good_rays"]

                        if not attributes is None:
                            for key in attributes.keys():
                                plot.attrs[key] = attributes[key]

            def add_attribute(self, attribute_name, attribute_value, dataset_name=None):
                if not dataset_name is None:
                    self[dataset_name].attrs[attribute_name] = attribute_value
                else:
                    self.attrs[attribute_name] = attribute_value

            def get_attribute(self,  attribute_name, dataset_name=None):
                if not dataset_name is None:
                    return self[dataset_name].attrs[attribute_name]
                else:
                    return self.attrs[attribute_name]

        class HistogramHdf5File(h5py.File):
            def __init__(self, file_name, mode="w"):
                super(ShadowPlot.HistogramHdf5File, self).__init__(name=file_name, mode=mode)

                self.coordinates = self.create_group("coordinates")
                self.plots = self.create_group("histogram_plots")
                self.last_plot = self.plots.create_group("last_plot")
                self.has_last_plot = False
                self.has_coordinate = False

                self.attrs["default"]          = "coordinates/X"
                self.attrs["file_name"]        = file_name
                self.attrs["file_time"]        = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
                self.attrs["creator"]          = "PlotXYHdf5File.__init__"
                self.attrs["code"]             = "ShadowOui"
                self.attrs["HDF5_Version"]     = h5py.version.hdf5_version
                self.attrs["h5py_version"]     = h5py.version.version

            def write_coordinates(self, ticket):
                if not self.has_coordinate:
                    self.x = self.coordinates.create_dataset("X", data=ticket["bin_center"])
                    self.has_coordinate = True
                else:
                    self.x[...] = ticket["bin_center"]

                self.coordinates.attrs["x_label"] = ShadowPlot.get_shadow_label(ticket["col"])

            def add_histogram(self, ticket, plot_name="last_plot", dataset_name="intensity", attributes={}):
                    if plot_name is None or plot_name.strip() == "" or plot_name.strip() == "last_plot":
                        if not self.has_last_plot:
                            self.lp_histogram  = self.last_plot.create_dataset(dataset_name, data=ticket['histogram'])
                            self.has_last_plot = True
                        else:
                            if self.lp_histogram.name != "/histogram_plots/last_plot/" + dataset_name:
                                self.last_plot.move(self.lp_histogram.name, "/histogram_plots/last_plot/" + dataset_name)

                            self.lp_histogram[...] = ticket['histogram']

                        self.last_plot.attrs["intensity"] = ticket["intensity"]
                        self.last_plot.attrs["total_rays"] = ticket["nrays"]
                        self.last_plot.attrs["good_rays"] = ticket["good_rays"]
                        self.last_plot.attrs["lost_rays"] = ticket["nrays"]-ticket["good_rays"]

                        if not attributes is None:
                            for key in attributes.keys():
                                self.last_plot.attrs[key] = attributes[key]
                    else:
                        plot = self.plots.create_group(plot_name)

                        plot.create_dataset(dataset_name, data=ticket['histogram'])
                        plot.attrs["intensity"]  = ticket["intensity"]
                        plot.attrs["total_rays"] = ticket["nrays"]
                        plot.attrs["good_rays"]  = ticket["good_rays"]
                        plot.attrs["lost_rays"]  = ticket["nrays"]-ticket["good_rays"]

                        if not attributes is None:
                            for key in attributes.keys():
                                plot.attrs[key] = attributes[key]
except:
    pass

class ShadowPreProcessor:

    @classmethod
    def read_surface_error_file(cls, filename):
        file = open(congruence.checkFile(filename), "r")

        rows = file.readlines()

        dimensions = rows[0].split()
        n_x = int(dimensions[0])
        n_y = int(dimensions[1])

        if n_x > 500:
            raise Exception("Malformed file: maximum allowed point in X direction is 500")

        x_coords = numpy.zeros(0)
        y_coords = numpy.zeros(0)
        z_values = numpy.zeros((n_x, n_y))


        index = 1
        dim_y_row = len(rows[index].split())
        is_ycoord = True
        first_x_row_index = 0

        while(is_ycoord):
            y_values = rows[index].split()

            if len(y_values) == dim_y_row:
                for y_value in y_values:
                    y_coords = numpy.append(y_coords, float(y_value))
            else:
                first_x_row_index = index
                is_ycoord = False

            index +=1

        first_x_row = rows[first_x_row_index].split()

        if len(first_x_row) == 2:
            x_index = 0
            z_index = 0

            for index in range(first_x_row_index, len(rows)):
                if z_index == 0:
                    values = rows[index].split()
                    x_coords = numpy.append(x_coords, float(values[0]))
                    z_value = float(values[1])
                else:
                    z_value = float(rows[index])

                z_values[x_index, z_index] = z_value
                z_index += 1

                if z_index == n_y:
                    x_index += 1
                    z_index = 0
        else:
            x_rows = []

            for index in range(2, len(rows)):

                x_row = rows[index].split("\t")

                if len(x_row) != 1 + n_y:
                    x_row = rows[index].split()

                if len(x_row) != 1 + n_y:
                    raise Exception("Malformed file: check format")

                x_rows.append(x_row)

            for x_index in range(0, len(x_rows)):
                x_coords = numpy.append(x_coords, float(x_rows[x_index][0]))

                for z_index in range(0, len(x_rows[x_index]) - 1):
                    z_value = float(x_rows[x_index][z_index + 1])

                    z_values[x_index, z_index] = z_value

        return x_coords, y_coords, z_values

class ShadowMath:

    @classmethod
    def gaussian_fit(cls, data_x, data_y):
        x = asarray(data_x)
        y = asarray(data_y)
        y_norm = y/sum(y)

        mean = sum(x*y_norm)
        sigma = numpy.sqrt(sum(y_norm*(x-mean)**2)/len(x))
        amplitude = max(y)

        parameters, covariance_matrix = optimize.curve_fit(ShadowMath.gaussian_function, x, y, p0 = [amplitude, mean, sigma])
        parameters.resize(4)
        parameters[3] = 2.355*parameters[2]# FWHM

        return parameters, covariance_matrix

    @classmethod
    def gaussian_function(cls, x, A, x0, sigma):
        return A*numpy.exp(-(x-x0)**2/(2*sigma**2))

    @classmethod
    def pseudovoigt_fit(cls, data_x, data_y):
        x = asarray(data_x)
        y = asarray(data_y)
        y_norm = y/sum(y)

        amplitude = max(data_y)
        mean = sum(x*y_norm)
        fwhm = numpy.sqrt(sum(y_norm*(x-mean)**2)/len(x))*2.355
        mixing = 0.1

        parameters, covariance_matrix = optimize.curve_fit(ShadowMath.pseudovoigt_function,
                                                           x, y,
                                                           p0 = [amplitude, mean, fwhm, mixing],
                                                           bounds = ([0.1*amplitude, 0.9*mean,  0.1*fwhm, 1e-3],
                                                                     [10.0*amplitude, 1.1*mean, 10.0*fwhm, 1.0]))

        return parameters, covariance_matrix

    @classmethod
    def pseudovoigt_function(cls, x, A, x0, fwhm, mixing):
        sigma = fwhm/2.355
        gamma = fwhm/2

        return A*(mixing*numpy.exp(-(x-x0)**2/(2*sigma**2)) + (1-mixing)*((gamma**2)/((x-x0)**2 + gamma**2)))

    @classmethod
    def caglioti_broadening_fit(cls, data_x, data_y):
        x = asarray(data_x)
        y = asarray(data_y)

        parameters, covariance_matrix = optimize.curve_fit(ShadowMath.caglioti_broadening_function,
                                                           x, y,
                                                           p0=[0.0001, 0.0001, 0.0001],
                                                           bounds = ([ -1.0, -1.0, -1.0],
                                                                     [ 1.0,  1.0,  1.0]))

        return parameters, covariance_matrix

    @classmethod
    def caglioti_broadening_function(cls, x, U, V, W):
        return numpy.sqrt(W + V * (numpy.tan(x*numpy.pi/360)) + U * (numpy.tan(x*numpy.pi/360))**2)

    @classmethod
    def caglioti_shape_fit(cls, data_x, data_y):
        x = asarray(data_x)
        y = asarray(data_y)

        parameters, covariance_matrix = optimize.curve_fit(ShadowMath.caglioti_shape_function,
                                                           x, y,
                                                           p0=[0.1, 0.0, 0.0],
                                                           bounds = ([ 0.0, -1.0, -1.0],
                                                                     [ 1.0,  1.0,  1.0]))

        return parameters, covariance_matrix

    @classmethod
    def caglioti_shape_function(cls, x, a, b, c):
        return a + b*(x*numpy.pi/360) + c*(x*numpy.pi/360)**2


    @classmethod
    def vectorial_product(cls, vector1, vector2):
        result = [0.0, 0.0, 0.0]

        result[0] = vector1[1]*vector2[2] - vector1[2]*vector2[1]
        result[1] = -(vector1[0]*vector2[2] - vector1[2]*vector2[0])
        result[2] = vector1[0]*vector2[1] - vector1[1]*vector2[0]

        return result

    @classmethod
    def scalar_product(cls, vector1, vector2):
        return vector1[0]*vector2[0] + vector1[1]*vector2[1] + vector1[2]*vector2[2]

    @classmethod
    def vector_modulus(cls, vector):
        return numpy.sqrt(cls.scalar_product(vector, vector))

    @classmethod
    def vector_multiply(cls, vector, constant):
        result = [0.0, 0.0, 0.0]

        result[0] = vector[0] * constant
        result[1] = vector[1] * constant
        result[2] = vector[2] * constant

        return result

    @classmethod
    def vector_divide(cls, vector, constant):
        result = [0.0, 0.0, 0.0]

        result[0] = vector[0] / constant
        result[1] = vector[1] / constant
        result[2] = vector[2] / constant

        return result

    @classmethod
    def vector_normalize(cls, vector):
        return cls.vector_divide(vector, cls.vector_modulus(vector))

    @classmethod
    def vector_sum(cls, vector1, vector2):
        result = [0.0, 0.0, 0.0]

        result[0] = vector1[0] + vector2[0]
        result[1] = vector1[1] + vector2[1]
        result[2] = vector1[2] + vector2[2]

        return result

    @classmethod
    def vector_difference(cls, vector1, vector2):
        result = [0.0, 0.0, 0.0]

        result[0] = vector1[0] - vector2[0]
        result[1] = vector1[1] - vector2[1]
        result[2] = vector1[2] - vector2[2]

        return result

    ##########################################################################
    # Rodrigues Formula:
    #
    # rotated = vector * cos(rotation_angle) + 
    #           (rotation_axis x vector) * sin(rotation_angle) + 
    #           rotation_axis*(rotation_axis . vector)(1 - cos(rotation_angle))
    #
    # rotation_angle in radians
    #
    ##########################################################################
    @classmethod
    def vector_rotate(cls, rotation_axis, rotation_angle, vector):

        result_temp_1 = ShadowMath.vector_multiply(vector, numpy.cos(rotation_angle))
        result_temp_2 = ShadowMath.vector_multiply(ShadowMath.vectorial_product(rotation_axis, vector),
                                                   numpy.sin(rotation_angle))
        result_temp_3 = ShadowMath.vector_multiply(ShadowMath.vector_multiply(rotation_axis, ShadowMath.scalar_product(rotation_axis, vector)), (1 - numpy.cos(rotation_angle)))

        result = ShadowMath.vector_sum(result_temp_1,
                                       ShadowMath.vector_sum(result_temp_2, result_temp_3))
        return result

    @classmethod
    def point_distance(cls, point1, point2):
        return cls.vector_modulus(cls.vector_difference(point1, point2))

class ShadowPhysics:

    A2EV = (codata.h*codata.c/codata.e)*1e+10
    K2EV = 2*numpy.pi/(codata.h*codata.c/codata.e*1e+2)

    @classmethod
    def getWavelengthFromShadowK(cls, k_mod): # in cm
        return (2*numpy.pi/k_mod)*1e+8 # in Angstrom

    @classmethod
    def getShadowKFromWavelength(cls, wavelength): # in A
        return (2*numpy.pi/wavelength)*1e+8 # in cm

    @classmethod
    def getWavelengthFromEnergy(cls, energy): #in eV
        return cls.A2EV/energy # in Angstrom

    @classmethod
    def getEnergyFromWavelength(cls, wavelength): # in Angstrom
        return cls.A2EV/wavelength # in eV

    @classmethod
    def getEnergyFromShadowK(cls, k_mod): # in cm
        return  k_mod/cls.K2EV # in eV

    @classmethod
    def getShadowKFromEnergy(cls, energy): # in A
        return cls.K2EV*energy # in cm

    @classmethod
    def calculateBraggAngle(cls, wavelength, h, k, l, a):

        # lambda = 2 pi / |k| = 2 d sen(th)
        #
        # sen(th) = lambda / (2  d)
        #
        # d = a / sqrt(h\u00b2 + k^2 + l^2)
        #
        # sen(th) = (sqrt(h^2 + k^2 + l^2) * lambda)/(2 a)

        theta_bragg = -1

        argument = wavelength*numpy.sqrt(h**2+k**2+l**2)/(2*a)
        if argument <= 1:
            result = numpy.arcsin(argument)
            if result > 0: theta_bragg = result

        return theta_bragg

    @classmethod
    def checkCompoundName(cls, compound_name):
        if compound_name is None: raise Exception("Compound Name is Empty")
        if str(compound_name.strip()) == "": raise Exception("Compound Name is Empty")

        compound_name = compound_name.strip()

        try:
            xraylib.CompoundParser(compound_name)
            return compound_name
        except:
            raise Exception("Compound Name is not correct")

    @classmethod
    def getMaterialDensity(cls, material_name):
        if material_name is None: return 0.0
        if str(material_name.strip()) == "": return 0.0

        try:
            compoundData = xraylib.CompoundParser(material_name)

            n_elements = compoundData["nElements"]
            if n_elements == 1:
                return xraylib.ElementDensity(compoundData["Elements"][0])
            else:
                density = 0.0
                mass_fractions = compoundData["massFractions"]
                elements = compoundData["Elements"]
                for i in range(n_elements): density += xraylib.ElementDensity(elements[i]) * mass_fractions[i]
                return density
        except:
            return 0.0

    @classmethod
    def ConstatoBackgroundNoised(cls, constant_value=0, n_sigma=1.0, random_generator=random.Random()):
        sigma = numpy.sqrt(constant_value) # poisson statistic
        noise = (n_sigma*sigma)*random_generator.random()
        sign_marker = random_generator.random()

        if sign_marker > 0.5:
            return int(round(constant_value+noise, 0))
        else:
            return int(round(constant_value-noise, 0))

    @classmethod
    def Chebyshev(cls, n, x):
        if n==0: return 1
        elif n==1: return x
        else: return 2*x*cls.Chebyshev(n-1, x)-cls.Chebyshev(n-2, x)

    @classmethod
    def ChebyshevBackground(cls, coefficients=[0,0,0,0,0,0], twotheta=0):
        coefficients_set = range(0, len(coefficients))
        background = 0

        for index in coefficients_set:
            background += coefficients[index]*cls.Chebyshev(index, twotheta)

        return background

    @classmethod
    def ChebyshevBackgroundNoised(cls, coefficients=[0,0,0,0,0,0], twotheta=0.0, n_sigma=1.0, random_generator=random.Random()):
        background = cls.ChebyshevBackground(coefficients, twotheta)
        sigma = numpy.sqrt(background) # poisson statistic

        noise = (n_sigma*sigma)*random_generator.random()
        sign_marker = random_generator.random()

        if sign_marker > 0.5:
            return int(round(background+noise, 0))
        else:
            return int(round(background-noise, 0))

    @classmethod
    def ExpDecay(cls, h, x):
      return numpy.exp(-h*x)

    @classmethod
    def ExpDecayBackground(cls, coefficients=[0,0,0,0,0,0], decayparams=[0,0,0,0,0,0], twotheta=0):
        coefficients_set = range(0, len(coefficients))
        background = 0

        for index in coefficients_set:
            background += coefficients[index]*cls.ExpDecay(decayparams[index], twotheta)

        return background

    @classmethod
    def ExpDecayBackgroundNoised(cls, coefficients=[0,0,0,0,0,0], decayparams=[0,0,0,0,0,0], twotheta=0, n_sigma=1, random_generator=random.Random()):
        background = cls.ExpDecayBackground(coefficients, decayparams, twotheta)
        sigma = numpy.sqrt(background) # poisson statistic

        noise = (n_sigma*sigma)*random_generator.random()
        sign_marker = random_generator.random()

        if sign_marker > 0.5:
            return int(round(background+noise, 0))
        else:
            return int(round(background-noise, 0))

import re
import time

class Properties(object):
    def __init__(self, props=None):
        self._props = {}
        self._origprops = {}
        self._keymap = {}
        self.othercharre = re.compile(r'(?<!\\)(\s*\=)|(?<!\\)(\s*\:)')
        self.othercharre2 = re.compile(r'(\s*\=)|(\s*\:)')
        self.bspacere = re.compile(r'\\(?!\s$)')

    def __str__(self):
        s='{'
        for key,value in self._props.items():
            s = ''.join((s,key,'=',value,', '))

        s=''.join((s[:-2],'}'))
        return s

    def __parse(self, lines):
        # Every line in the file must consist of either a comment
        # or a key-value pair. A key-value pair is a line consisting
        # of a key which is a combination of non-white space characters
        # The separator character between key-value pairs is a '=',
        # ':' or a whitespace character not including the newline.
        # If the '=' or ':' characters are found, in the line, even
        # keys containing whitespace chars are allowed.
        # A line with only a key according to the rules above is also
        # fine. In such case, the value is considered as the empty string.
        # In order to include characters '=' or ':' in a key or value,
        # they have to be properly escaped using the backslash character.

        # Some examples of valid key-value pairs:
        #
        # key     value
        # key=value
        # key:value
        # key     value1,value2,value3
        # key     value1,value2,value3 \
        #         value4, value5
        # key
        # This key= this value
        # key = value1 value2 value3

        # Any line that starts with a '#' is considered a comment
        # and skipped. Also any trailing or preceding whitespaces
        # are removed from the key/value.

        # This is a line parser. It parses the
        # contents like by line.

        lineno=0
        i = iter(lines)

        for line in i:
            lineno += 1
            line = line.strip()

            if not line: continue
            if line[0] == '#': continue

            sepidx = -1

            m = self.othercharre.search(line)
            if m:
                first, last = m.span()
                start, end = 0, first
                wspacere = re.compile(r'(?<![\\\=\:])(\s)')
            else:
                if self.othercharre2.search(line):
                    wspacere = re.compile(r'(?<![\\])(\s)')

                start, end = 0, len(line)

            m2 = wspacere.search(line, start, end)
            if m2:
                first, last = m2.span()
                sepidx = first
            elif m:
                first, last = m.span()
                sepidx = last - 1

            while line[-1] == '\\':
                nextline = i.next()
                nextline = nextline.strip()
                lineno += 1
                line = line[:-1] + nextline

            if sepidx != -1:
                key, value = line[:sepidx], line[sepidx+1:]
            else:
                key,value = line,''

            self.processPair(key, value)

    def processPair(self, key, value):
        oldkey = key
        oldvalue = value

        keyparts = self.bspacere.split(key)

        strippable = False
        lastpart = keyparts[-1]

        if lastpart.find('\\ ') != -1:
            keyparts[-1] = lastpart.replace('\\','')

        elif lastpart and lastpart[-1] == ' ':
            strippable = True

        key = ''.join(keyparts)
        if strippable:
            key = key.strip()
            oldkey = oldkey.strip()

        oldvalue = self.unescape(oldvalue)
        value = self.unescape(value)

        self._props[key] = value.strip()

        if self._keymap.__contains__(key):
            oldkey = self._keymap.get(key)
            self._origprops[oldkey] = oldvalue.strip()
        else:
            self._origprops[oldkey] = oldvalue.strip()
            self._keymap[key] = oldkey

    def escape(self, value):
        newvalue = value.replace(':','\:')
        newvalue = newvalue.replace('=','\=')

        return newvalue

    def unescape(self, value):
        newvalue = value.replace('\:',':')
        newvalue = newvalue.replace('\=','=')

        return newvalue

    def load(self, stream):
        if not hasattr(stream, 'read'):
            raise TypeError('Argument should be a file object!')
        if stream.mode != 'r':
            raise ValueError ('Stream should be opened in read-only mode!')

        try:
            lines = stream.readlines()
            self.__parse(lines)
        except IOError as e:
            raise e

    def getProperty(self, key):
        return self._props.get(key)

    def setProperty(self, key, value):
        if type(key) is str and type(value) is str:
            self.processPair(key, value)
        else:
            raise TypeError('both key and value should be strings!')

    def propertyNames(self):
        return self._props.keys()

    def list(self, out=sys.stdout):
        out.write('-- listing properties --\n')
        for key,value in self._props.items():
            out.write(''.join((key,'=',value,'\n')))

    def store(self, out, header=""):
        if out.mode[0] != 'w':
            raise ValueError('Steam should be opened in write mode!')

        try:
            out.write(''.join(('#',header,'\n')))
            tstamp = time.strftime('%a %b %d %H:%M:%S %Z %Y', time.localtime())
            out.write(''.join(('#',tstamp,'\n')))
            for prop, val in self._origprops.items():
                out.write(''.join((prop,'=',self.escape(val),'\n')))

            out.close()
        except IOError as e:
            raise e

    def getPropertyDict(self):
        return self._props

    def __getitem__(self, name):
        return self.getProperty(name)

    def __setitem__(self, name, value):
        self.setProperty(name, value)

    def __getattr__(self, name):
        try:
            return self.__dict__[name]
        except KeyError:
            if hasattr(self._props,name):
                return getattr(self._props, name)

try:
    from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
    from matplotlib.figure import Figure
    from mpl_toolkits.mplot3d import Axes3D  # necessario per caricare i plot 3D
except:
    pass

try:
    from PyQt5.QtWidgets import QApplication, QVBoxLayout
    from PyQt5.QtCore import QCoreApplication

    class MathTextLabel(QWidget):
        def __init__(self, mathText, size=None, parent=None, **kwargs):
            QWidget.__init__(self, parent, **kwargs)

            l=QVBoxLayout(self)
            l.setContentsMargins(0,0,0,0)

            r,g,b,a=self.palette().base().color().getRgbF()

            self._figure=Figure(edgecolor=(r,g,b), facecolor=(r,g,b))
            self._canvas=FigureCanvas(self._figure)
            l.addWidget(self._canvas)

            self._figure.clear()

            if not size:
                size = QCoreApplication.instance().font().pointSize()

            text=self._figure.suptitle(
                mathText,
                x=0.0,
                y=1.0,
                horizontalalignment='left',
                verticalalignment='top',
                size=size)
            self._canvas.draw()

            (x0,y0),(x1,y1)=text.get_window_extent().get_points()
            w=x1-x0; h=y1-y0

            self._figure.set_size_inches(w/80, h/80)
            self.setFixedSize(w,h)
except:
    pass


if __name__ == "__main__":

    #print(congruence.checkFileName("pippo.dat"))
    #print(congruence.checkFileName("Files/pippo.dat"))
    #print(congruence.checkFileName("Files/pippo.dat"))
    #print(congruence.checkFileName("/Users/labx/Desktop/pippo.dat"))

    s = " 5   8095683980.2420149       3.34799999999999994E-008"

    print(s.strip().split(" "))


    print("Bragg")
    ShadowCongruence.checkBraggFile("/Users/labx/Oasys/bragg.dat")
    print("PreRefl")
    ShadowCongruence.checkPreReflFile("/Users/labx/Oasys/reflec.dat")
    print("PreMLayer")
    ShadowCongruence.checkPreMLayerFile("/Users/labx/Oasys/mlayer.dat")

    #ShadowCongruence.checkXOPDiffractionProfileFile("/Users/labx/Oasys/mlayer.dat")

    '''
    print(ShadowPhysics.A2EV)

    print(ShadowPhysics.Chebyshev(4, 21))
    print(ShadowPhysics.Chebyshev(0, 35))

    coefficients = [5.530814e+002, 2.487256e+000, -2.004860e-001, 2.246427e-003, -1.044517e-005, 1.721576e-008]
    random_generator=random.Random()

    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 10, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 11, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 12, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 13, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 14, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 15, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 16, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 17, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 18, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 19, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 20, random_generator=random_generator))
    '''
    '''
    import matplotlib.pyplot as plt

    x_coords, y_coords, z_values = ShadowPreProcessor.read_surface_error_file("/Users/labx/Oasys/mirror.dat")

    fig = plt.figure()
    ax = fig.gca(projection='3d')
    X, Y = numpy.meshgrid(x_coords, y_coords)
    surf = ax.plot_surface(X, Y, z_values.T, rstride=1, cstride=1, cmap=cm.coolwarm,
            linewidth=0, antialiased=False)
    #ax.set_zlim(-1.01, 1.01)

    fig.colorbar(surf, shrink=0.5, aspect=5)

    plt.show()

    app = QApplication(sys.argv)

    widget = QWidget()
    widget.setLayout(QVBoxLayout())

    figure = Figure(figsize=(100, 100))
    figure.patch.set_facecolor('white')

    axis = figure.add_subplot(111, projection='3d')

    axis.set_xlabel("X (cm)")
    axis.set_ylabel("Y (cm)")
    axis.set_zlabel("Z (cm)")

    figure_canvas = FigureCanvasQTAgg(figure)
    widget.layout().addWidget(figure_canvas)
    figure_canvas.setFixedWidth(500)
    figure_canvas.setFixedHeight(450)

    x_coords, y_coords, z_values = ShadowPreProcessor.read_surface_error_file("/Users/labx/Oasys/mirror.dat")

    x_to_plot, y_to_plot = numpy.meshgrid(x_coords, y_coords)

    axis.plot_surface(x_to_plot, y_to_plot, z_values.T,
                      rstride=1, cstride=1, cmap=cm.autumn, linewidth=0.5, antialiased=True)

    figure_canvas.draw()
    figure_canvas.show()
    widget.show()

    app.exec()
    '''



import sys
import time
import numpy

from orangewidget import gui
from orangewidget.settings import Setting

from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence
from oasys.util.oasys_util import EmittingStream

from orangecontrib.shadow4.widgets.gui.ow_electron_beam import OWElectronBeam
from orangecontrib.shadow4.util.shadow4_objects import ShadowData


from syned.beamline.beamline import Beamline
from shadow4.beamline.s4_beamline import S4Beamline

from syned.storage_ring.magnetic_structures.bending_magnet import BendingMagnet
from syned.widget.widget_decorator import WidgetDecorator

from shadow4.sources.bending_magnet.s4_bending_magnet import S4BendingMagnet
from shadow4.sources.bending_magnet.s4_bending_magnet_light_source import S4BendingMagnetLightSource

class OWBendingMagnet(OWElectronBeam, WidgetDecorator):

    name = "Bending Magnet"
    description = "Shadow Source: Bending Magnet"
    icon = "icons/bending_magnet.png"
    priority = 3

    inputs = []
    WidgetDecorator.append_syned_input_data(inputs)

    outputs = [{"name":"Shadow Data",
                "type":ShadowData,
                "doc":"",}]


    number_of_rays = Setting(5000)
    seed = Setting(5676561)


    magnetic_field         = Setting(-1.26754)
    divergence             = Setting(69e-3)
    emin                   = Setting(1000.0)  # Photon energy scan from energy (in eV)
    emax                   = Setting(1000.1)  # Photon energy scan to energy (in eV)
    ng_e                   = Setting(100)     # Photon energy scan number of points
    ng_j                   = Setting(100) # Number of points in electron trajectory (per period) for internal calculation only


    def __init__(self):
        super().__init__()

        tab_bas = oasysgui.createTabPage(self.tabs_control_area, "Bending Magnet Setting")

        #
        box_1 = oasysgui.widgetBox(tab_bas, "Bending Magnet Parameters", addSpace=True, orientation="vertical")
        oasysgui.lineEdit(box_1, self, "emin", "Minimum Energy [eV]", tooltip="emin", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_1, self, "emax", "Maximum Energy [eV]", tooltip="emax", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_1, self, "magnetic_field", "Magnetic Field [T]", tooltip="magnetic_field", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_1, self, "divergence", "Horizontal divergence (arc of radius) [rads]", tooltip="divergence", labelWidth=260, valueType=float, orientation="horizontal")

        box_2 = oasysgui.widgetBox(tab_bas, "Sampling rays", addSpace=True, orientation="vertical")
        oasysgui.lineEdit(box_2, self, "number_of_rays", "Number of Rays", tooltip="Number of Rays", labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(box_2, self, "seed", "Seed", tooltip="Seed (0=clock)", labelWidth=250, valueType=int, orientation="horizontal")

        box_3 = oasysgui.widgetBox(tab_bas, "Optional parameters for internal calculation", addSpace=True, orientation="vertical")
        oasysgui.lineEdit(box_3, self, "ng_e", "Spectrum number of points", tooltip="ng_e", labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(box_3, self, "ng_j", "Electron trajectory number of points", tooltip="ng_j", labelWidth=250, valueType=int, orientation="horizontal")

        gui.rubber(self.controlArea)
        gui.rubber(self.mainArea)


    def checkFields(self):
        self.number_of_rays = congruence.checkPositiveNumber(self.number_of_rays, "Number of rays")
        self.seed = congruence.checkPositiveNumber(self.seed, "Seed")
        # self.energy = congruence.checkPositiveNumber(self.energy, "Energy")
        # self.delta_e = congruence.checkPositiveNumber(self.delta_e, "Delta Energy")
        # self.undulator_length = congruence.checkPositiveNumber(self.undulator_length, "Undulator Length")

    def get_lightsource(self):
        # syned electron beam
        electron_beam = self.get_electron_beam()
        print("\n\n>>>>>> electron_beam info: ", electron_beam.info())

        if self.type_of_properties == 3:
            flag_emittance = 0
        else:
            flag_emittance = 1

        magnetic_radius = S4BendingMagnet.calculate_magnetic_radius(self.magnetic_field, electron_beam.energy())
        length = numpy.abs(self.divergence * magnetic_radius)

        print(">>> calculated magnetic_radius = S4BendingMagnet.calculate_magnetic_radius(%f, %f) = %f" %\
              (self.magnetic_field, electron_beam.energy(), magnetic_radius))

        print(">>> calculated BM length = divergence * magnetic_radius = %f " % length)

        bm = S4BendingMagnet(magnetic_radius,self.magnetic_field,length,
                             emin=self.emin,  # Photon energy scan from energy (in eV)
                             emax=self.emax,  # Photon energy scan to energy (in eV)
                             ng_e=self.ng_e,  # Photon energy scan number of points
                             ng_j=self.ng_j,
                             # Number of points in electron trajectory (per period) for internal calculation only
                             flag_emittance=flag_emittance,  # when sampling rays: Use emittance (0=No, 1=Yes)
                             )


        # bm = S4BendingMagnet.initialize_from_magnetic_field_divergence_and_electron_energy(
        #     magnetic_field=self.magnetic_field,
        #     divergence=self.divergence,
        #     electron_energy_in_GeV=electron_beam.energy(),
        #     emin=self.emin,# Photon energy scan from energy (in eV)
        #     emax=self.emax,# Photon energy scan to energy (in eV)
        #     ng_e=self.ng_e,# Photon energy scan number of points
        #     ng_j=self.ng_j,# Number of points in electron trajectory (per period) for internal calculation only
        #     flag_emittance=flag_emittance,# when sampling rays: Use emittance (0=No, 1=Yes)
        #     )


        print("\n\n>>>>>> BM info: ", bm.info())


        # S4UndulatorLightSource
        lightsource = S4BendingMagnetLightSource(name='BendingMagnet',
                                             electron_beam=electron_beam,
                                             magnetic_structure=bm,
                                             nrays=self.number_of_rays,
                                             seed=self.seed)

        print("\n\n>>>>>> S4BendingMagnetLightSource info: ", lightsource.info())

        return lightsource

    def run_shadow4(self):

        sys.stdout = EmittingStream(textWritten=self._write_stdout)

        self._set_plot_quality()

        self.progressBarInit()

        light_source = self.get_lightsource()

        self.progressBarSet(5)


        # run shadow4

        t00 = time.time()
        print(">>>> starting calculation...")
        output_beam = light_source.get_beam()
        # todo:
        # photon_energy, flux, spectral_power = light_source.calculate_spectrum()
        t11 = time.time() - t00
        print(">>>> time for %d rays: %f s, %f min, " % (self.number_of_rays, t11, t11 / 60))


        #
        # beam plots
        #
        self._plot_results(output_beam, None, progressBarValue=80)

        #
        # script
        #
        script = light_source.to_python_code()
        script += "\n\n# test plot\nfrom srxraylib.plot.gol import plot_scatter"
        script += "\nrays = beam.get_rays()"
        script += "\nplot_scatter(1e6 * rays[:, 0], 1e6 * rays[:, 2], title='(X,Z) in microns')"
        self.shadow4_script.set_code(script)

        self.progressBarFinished()

        #
        # send beam
        #
        self.send("Shadow Data", ShadowData(beam=output_beam,
                                           number_of_rays=self.number_of_rays,
                                           beamline=S4Beamline(light_source=light_source)))


    def receive_syned_data(self, data):
        if data is not None:
            if isinstance(data, Beamline):
                if not data.get_light_source() is None:
                    if isinstance(data.get_light_source().get_magnetic_structure(), BendingMagnet):
                        light_source = data.get_light_source()

                        # self.energy =  round(light_source.get_magnetic_structure().resonance_energy(light_source.get_electron_beam().gamma()), 3)
                        # self.delta_e = 0.0
                        # self.undulator_length = light_source.get_magnetic_structure().length()

                        self.populate_fields_from_syned_electron_beam(light_source.get_electron_beam())

                    else:
                        raise ValueError("Syned light source not congruent")
                else:
                    raise ValueError("Syned data not correct: light source not present")
            else:
                raise ValueError("Syned data not correct")

if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication
    a = QApplication(sys.argv)
    ow = OWBendingMagnet()
    ow.show()
    a.exec_()
    ow.saveSettings()



import sys
import time


from orangewidget import gui
from orangewidget.settings import Setting

from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence
from oasys.util.oasys_util import EmittingStream

from orangecontrib.shadow4.widgets.gui.ow_electron_beam import OWElectronBeam
from orangecontrib.shadow4.util.shadow4_objects import ShadowData


from syned.beamline.beamline import Beamline
from shadow4.beamline.s4_beamline import S4Beamline

from syned.storage_ring.magnetic_structures.undulator import Undulator
from syned.widget.widget_decorator import WidgetDecorator

from shadow4.sources.undulator.s4_undulator import S4Undulator
from shadow4.sources.undulator.s4_undulator_light_source import S4UndulatorLightSource

class OWUndulatorGaussian(OWElectronBeam, WidgetDecorator):

    name = "Undulator Gaussian"
    description = "Shadow Source: Undulator Gaussian"
    icon = "icons/ugaussian.png"
    priority = 5

    inputs = []
    WidgetDecorator.append_syned_input_data(inputs)

    outputs = [{"name":"Shadow Data",
                "type":ShadowData,
                "doc":"",}]

    undulator_length = Setting(4.0)
    energy = Setting(15000.0)
    delta_e = Setting(1500.0)
    number_of_rays = Setting(5000)
    seed = Setting(5676561)


    def __init__(self):
        super().__init__()

        tab_bas = oasysgui.createTabPage(self.tabs_control_area, "Undulator Setting")

        #
        box_1 = oasysgui.widgetBox(tab_bas, "Undulator Parameters", addSpace=True, orientation="vertical")
        oasysgui.lineEdit(box_1, self, "undulator_length", "Undulator Length [m]", labelWidth=250, tooltip="Undulator Length [m]", valueType=float, orientation="horizontal")
        #
        box_2 = oasysgui.widgetBox(tab_bas, "Sampling rays", addSpace=True, orientation="vertical")
        oasysgui.lineEdit(box_2, self, "energy", "Set resonance at energy [eV]", tooltip="Set resonance at energy [eV]", labelWidth=250, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_2, self, "delta_e", "Delta Energy [eV]", tooltip="Delta Energy [eV]", labelWidth=250, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_2, self, "number_of_rays", "Number of Rays", tooltip="Number of Rays", labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(box_2, self, "seed", "Seed", tooltip="Seed (0=clock)", labelWidth=250, valueType=int, orientation="horizontal")

        gui.rubber(self.controlArea)
        gui.rubber(self.mainArea)


    def checkFields(self):
        self.number_of_rays = congruence.checkPositiveNumber(self.number_of_rays, "Number of rays")
        self.seed = congruence.checkPositiveNumber(self.seed, "Seed")
        self.energy = congruence.checkPositiveNumber(self.energy, "Energy")
        self.delta_e = congruence.checkPositiveNumber(self.delta_e, "Delta Energy")
        self.undulator_length = congruence.checkPositiveNumber(self.undulator_length, "Undulator Length")

    def get_lightsource(self):
        # syned electron beam
        electron_beam = self.get_electron_beam()
        print("\n\n>>>>>> electron_beam info: ", electron_beam.info())

        if self.type_of_properties == 3:
            flag_emittance = 0
        else:
            flag_emittance = 1

        sourceundulator = S4Undulator(
            # K_vertical=1.0,  # syned Undulator parameter
            period_length=self.undulator_length/100,  # syned Undulator parameter
            number_of_periods=100,  # syned Undulator parameter
            emin=self.energy - 0.5 * self.delta_e,  # Photon energy scan from energy (in eV)
            emax=self.energy + 0.5 * self.delta_e,  # Photon energy scan to energy (in eV)
            # ng_e=11,  # Photon energy scan number of points
            # maxangle=50e-6,  # Maximum radiation semiaperture in RADIANS
            # ng_t=31,  # Number of points in angle theta
            # ng_p=21,  # Number of points in angle phi
            # ng_j=20,  # Number of points in electron trajectory (per period) for internal calculation only
            # code_undul_phot="internal",  # internal, pysru, srw
            flag_emittance=flag_emittance,  # when sampling rays: Use emittance (0=No, 1=Yes)
            # flag_size=0,  # when sampling rays: 0=point,1=Gaussian,2=FT(Divergences)
            use_gaussian_approximation=1,
        )

        print("\n\n>>>>>> Undulator info: ", sourceundulator.info())

        if self.delta_e == 0:
            sourceundulator.set_energy_monochromatic(self.energy)
        else:
            sourceundulator.set_energy_box(self.energy-0.5*self.delta_e, self.energy+0.5*self.delta_e,)

        # S4UndulatorLightSource
        lightsource = S4UndulatorLightSource(name='GaussianUndulator',
                                             electron_beam=electron_beam,
                                             magnetic_structure=sourceundulator,
                                             nrays=self.number_of_rays,
                                             seed=self.seed)

        print("\n\n>>>>>> S4UndulatorLightSource info: ", lightsource.info())

        return lightsource

    def run_shadow4(self):

        sys.stdout = EmittingStream(textWritten=self._write_stdout)

        self._set_plot_quality()

        self.progressBarInit()

        light_source = self.get_lightsource()

        self.progressBarSet(5)


        # run shadow4

        t00 = time.time()
        print(">>>> starting calculation...")
        output_beam = light_source.get_beam()
        t11 = time.time() - t00
        print(">>>> time for %d rays: %f s, %f min, " % (self.number_of_rays, t11, t11 / 60))

        #
        # beam plots
        #
        self._plot_results(output_beam, None, progressBarValue=80)

        #
        # script
        #
        script = light_source.to_python_code()
        # script += "\n\n\n# run shadow4"
        # script += "\nbeam = light_source.get_beam(NRAYS=%d, SEED=%d)" % (self.number_of_rays, self.seed)
        script += "\n\n# test plot\nfrom srxraylib.plot.gol import plot_scatter"
        script += "\nrays = beam.get_rays()"
        script += "\nplot_scatter(1e6 * rays[:, 0], 1e6 * rays[:, 2], title='(X,Z) in microns')"
        self.shadow4_script.set_code(script)

        self.progressBarFinished()

        #
        # send beam
        #
        self.send("Shadow Data", ShadowData(beam=output_beam,
                                           number_of_rays=self.number_of_rays,
                                           beamline=S4Beamline(light_source=light_source)))


    def receive_syned_data(self, data):
        if data is not None:
            if isinstance(data, Beamline):
                if not data.get_light_source() is None:
                    if isinstance(data.get_light_source().get_magnetic_structure(), Undulator):
                        light_source = data.get_light_source()

                        self.energy =  round(light_source.get_magnetic_structure().resonance_energy(light_source.get_electron_beam().gamma()), 3)
                        self.delta_e = 0.0
                        self.undulator_length = light_source.get_magnetic_structure().length()

                        self.populate_fields_from_electron_beam(light_source.get_electron_beam())

                    else:
                        raise ValueError("Syned light source not congruent")
                else:
                    raise ValueError("Syned data not correct: light source not present")
            else:
                raise ValueError("Syned data not correct")

if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication
    a = QApplication(sys.argv)
    ow = OWUndulatorGaussian()
    ow.show()
    a.exec_()
    ow.saveSettings()



from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import QRect

from orangewidget import gui
from oasys.widgets import gui as oasysgui

from oasys.widgets.widget import AutomaticWidget
from orangewidget.settings import Setting

try:
    from orangecontrib.shadow.util.shadow_objects import ShadowBeam as ShadowBeam3
except:
    pass

from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from shadow4.beam.s4_beam import S4Beam

from shadow4.beamline.s4_beamline import S4Beamline

class OW_beam_converter_3_to_4(AutomaticWidget):
    name = "shadow3->4 beam converter"
    id = "toShadowOUIbeam"
    description = "shadow4->3 beam converter"
    icon = "icons/beam3to4.png"
    priority = 20
    category = ""
    keywords = ["shadow3", "shadow4"]

    inputs = [("Shadow3 Beam", ShadowBeam3, "set_input")]

    outputs = [{"name":"Shadow Data",
                "type":ShadowData,
                "doc":"",}]

    MAX_WIDTH = 420
    MAX_HEIGHT = 230
    CONTROL_AREA_WIDTH = 410

    want_main_area = 0

    pixels_h = Setting(100)
    pixels_v = Setting(100)

    shadow_beam = None

    def __init__(self):
        super().__init__()

        geom = QApplication.desktop().availableGeometry()
        self.setGeometry(QRect(round(geom.width()*0.05),
                               round(geom.height()*0.05),
                               round(min(geom.width()*0.98, self.MAX_WIDTH)),
                               round(min(geom.height()*0.95, self.MAX_HEIGHT))))

        self.setMinimumHeight(self.geometry().height())
        self.setMinimumWidth(self.geometry().width())
        self.setMaximumHeight(self.geometry().height())
        self.setMaximumWidth(self.geometry().width())

        self.controlArea.setFixedWidth(self.MAX_WIDTH-10)
        self.controlArea.setFixedHeight(self.MAX_HEIGHT-10)

        main_box = oasysgui.widgetBox(self.controlArea, "From Shadow3 (ShadowOUI) Beam To Shadow4 Beam", orientation="vertical", width=self.CONTROL_AREA_WIDTH-5, height=140)

        gui.button(main_box, self, "Compute", callback=self.convert_beam, height=45)

    def set_input(self, input_data):
        self.setStatusMessage("")

        if not input_data is None:
            self.shadow_beam = input_data
            if self.is_automatic_execution: self.convert_beam()

    def convert_beam(self):
        beam3 = self.shadow_beam._beam
        beam4 = ShadowData(beam=S4Beam(array=beam3.rays), beamline=S4Beamline())
        beam4.beam.rays[:, 0:3] *= self.workspace_units_to_m
        beam4.beam.rays[:, 12]  *= self.workspace_units_to_m

        print(beam4)
        self.send("Shadow Data", beam4)

if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication
    import sys
    import Shadow
    a = QApplication(sys.argv)
    ow = OW_beam_converter_3_to_4()
    ow.workspace_units_to_m = 1.0
    ow.set_input(ShadowBeam3(beam=Shadow.Beam(N=5000)))
    ow.show()
    a.exec_()
    #ow.saveSettings()




import sys
import numpy

from PyQt5.QtGui import QTextCursor
from PyQt5.QtWidgets import QApplication
from orangewidget import gui
from orangewidget.settings import Setting
from oasys.widgets import gui as oasysgui
from oasys.util.oasys_util import TriggerIn

from shadow4.beam.s4_beam import S4Beam
from orangecontrib.shadow4.widgets.gui.ow_automatic_element import AutomaticElement
from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from orangecontrib.shadow4.util.shadow4_util import ShadowPlot, ShadowCongruence
from orangecontrib.shadow4.util.python_script import PythonScript

class GenericElement(AutomaticElement):
    IMAGE_WIDTH  = 860
    IMAGE_HEIGHT = 545

    view_type = Setting(0)

    plotted_beam   = None
    footprint_beam = None
    has_footprint  = True

    def __init__(self, show_automatic_box=True, has_footprint=True):
        super().__init__(show_automatic_box)
        self.has_footprint = has_footprint

        self.main_tabs = oasysgui.tabWidget(self.mainArea)
        plot_tab = oasysgui.createTabPage(self.main_tabs, "Plots")
        out_tab = oasysgui.createTabPage(self.main_tabs, "Output")

        view_box = oasysgui.widgetBox(plot_tab, "Plotting Style", addSpace=False, orientation="horizontal")
        view_box_1 = oasysgui.widgetBox(view_box, "", addSpace=False, orientation="vertical", width=350)

        self.view_type_combo = gui.comboBox(view_box_1, self, "view_type", label="Select level of Plotting",
                                            labelWidth=220,
                                            items=["Detailed Plot", "Preview", "None"],
                                            callback=self._set_plot_quality, sendSelectedValue=False, orientation="horizontal")


        # script tab
        script_tab = oasysgui.createTabPage(self.main_tabs, "Script")
        self.shadow4_script = PythonScript()
        self.shadow4_script.code_area.setFixedHeight(400)

        script_box = gui.widgetBox(script_tab, "Python script", addSpace=True, orientation="horizontal")
        script_box.layout().addWidget(self.shadow4_script)

        self.tab = []
        self.tabs = oasysgui.tabWidget(plot_tab)

        self._initialize_tabs()

        self.shadow_output = oasysgui.textArea(height=580, width=800)

        out_box = gui.widgetBox(out_tab, "System Output", addSpace=True, orientation="horizontal")
        out_box.layout().addWidget(self.shadow_output)

    def _initialize_tabs(self):
        current_tab = self.tabs.currentIndex()

        size = len(self.tab)
        indexes = range(0, size)
        for index in indexes: self.tabs.removeTab(size-1-index)

        titles = self._get_titles()
        if self.has_footprint: self.plot_canvas = [None]*(len(titles) + 1)
        else:                  self.plot_canvas = [None]*len(titles)
        self.tab = []

        for title in titles: self.tab.append(oasysgui.createTabPage(self.tabs, title))
        if self.has_footprint: self.tab.append(gui.createTabPage(self.tabs, "Footprint"))

        for tab in self.tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)

        self.tabs.setCurrentIndex(current_tab)

    def _check_not_interactive_conditions(self, input_data : ShadowData):
        not_interactive = False

        if not (input_data is None or input_data.scanning_data is None):
            not_interactive = input_data.scanning_data.has_additional_parameter("total_power")

        return not_interactive

    def _send_empty_beam(self):
        empty_beam      = S4Beam(N=1)
        empty_beam.rays = numpy.array([])

        self.send("Shadow Data", ShadowData(beam=empty_beam))
        self.send("Trigger",    TriggerIn(new_object=True))

    def _set_plot_quality(self):
        self.progressBarInit()

        if not self.plotted_beam is None:
            try:
                self._initialize_tabs()
                self._plot_results(self.plotted_beam, self.footprint_beam, progressBarValue=80)
            except Exception as exception:
                self.prompt_exception(exception)

        self.progressBarFinished()

    def _plot_xy_preview(self, beam, progressBarValue, var_x, var_y, plot_canvas_index, title, xtitle, ytitle, is_footprint=False):
        if self.plot_canvas[plot_canvas_index] is None:
            self.plot_canvas[plot_canvas_index] = oasysgui.plotWindow(roi=False, control=False, position=True)
            self.plot_canvas[plot_canvas_index].setDefaultPlotLines(False)
            self.plot_canvas[plot_canvas_index].setActiveCurveColor(color='blue')

            self.tab[plot_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])

        ShadowPlot.plotxy_preview(self.plot_canvas[plot_canvas_index], beam, var_x, var_y, nolost=1, title=title, xtitle=xtitle, ytitle=ytitle, is_footprint=is_footprint)

        self.progressBarSet(progressBarValue)

    def _plot_xy_detailed(self, beam, progressBarValue, var_x, var_y, plot_canvas_index, title, xtitle, ytitle, xum="", yum="", is_footprint=False):
        if self.plot_canvas[plot_canvas_index] is None:
            self.plot_canvas[plot_canvas_index] = ShadowPlot.DetailedPlotWidget()
            self.tab[plot_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])

        self.plot_canvas[plot_canvas_index].plot_xy(beam, var_x, var_y, title, xtitle, ytitle, xum=xum, yum=yum, is_footprint=is_footprint)

        self.progressBarSet(progressBarValue)

    def _plot_histo_preview(self, beam, progressBarValue, var, plot_canvas_index, title, xtitle, ytitle):
        if self.plot_canvas[plot_canvas_index] is None:
            self.plot_canvas[plot_canvas_index] = oasysgui.plotWindow(roi=False, control=False, position=True)
            self.plot_canvas[plot_canvas_index].setDefaultPlotLines(True)
            self.plot_canvas[plot_canvas_index].setActiveCurveColor(color='blue')

            self.tab[plot_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])

        ShadowPlot.plot_histo_preview(self.plot_canvas[plot_canvas_index], beam, var, 1, 23, title, xtitle, ytitle)

        self.progressBarSet(progressBarValue)

    def _plot_histo_detailed(self, beam, progressBarValue, var, plot_canvas_index, title, xtitle, ytitle, xum=""):
        if self.plot_canvas[plot_canvas_index] is None:
            self.plot_canvas[plot_canvas_index] = ShadowPlot.DetailedHistoWidget()
            self.tab[plot_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])

        self.plot_canvas[plot_canvas_index].plot_histo(beam, var, 1, None, 23, title, xtitle, ytitle, xum=xum)

        self.progressBarSet(progressBarValue)

    def _plot_results(self, output_beam, footprint, progressBarValue=80):
        if not self.view_type == 2:
            if ShadowCongruence.check_empty_beam(output_beam):
                self.view_type_combo.setEnabled(False)

                ShadowPlot.set_conversion_active(self._is_conversion_active())

                variables = self._get_variables_to_plot()
                titles    = self._get_titles()
                xtitles   = self._get_x_titles()
                ytitles   = self._get_y_titles()
                xums      = self._get_x_um()
                yums      = self._get_y_um()

                try:
                    if self.view_type == 1:
                        self._plot_xy_preview(output_beam, progressBarValue + 4, variables[0][0], variables[0][1], plot_canvas_index=0, title=titles[0], xtitle=xtitles[0], ytitle=ytitles[0])
                        self._plot_xy_preview(output_beam, progressBarValue + 8, variables[1][0], variables[1][1], plot_canvas_index=1, title=titles[1], xtitle=xtitles[1], ytitle=ytitles[1])
                        self._plot_xy_preview(output_beam, progressBarValue + 12, variables[2][0], variables[2][1], plot_canvas_index=2, title=titles[2], xtitle=xtitles[2], ytitle=ytitles[2])
                        self._plot_xy_preview(output_beam, progressBarValue + 16, variables[3][0], variables[3][1], plot_canvas_index=3, title=titles[3], xtitle=xtitles[3], ytitle=ytitles[3])
                        self._plot_histo_preview(output_beam, progressBarValue + 20, variables[4], plot_canvas_index=4, title=titles[4], xtitle=xtitles[4], ytitle=ytitles[4])
                        if self.has_footprint: self._plot_xy_preview(footprint, progressBarValue + 20, 2, 1, plot_canvas_index=5, title="Footprint", xtitle="Y [m]", ytitle="X [m]", is_footprint=True)


                    elif self.view_type == 0:
                        self._plot_xy_detailed(output_beam, progressBarValue + 4, variables[0][0], variables[0][1], plot_canvas_index=0, title=titles[0], xtitle=xtitles[0], ytitle=ytitles[0], xum=xums[0], yum=yums[0])
                        self._plot_xy_detailed(output_beam, progressBarValue + 8, variables[1][0], variables[1][1], plot_canvas_index=1, title=titles[1], xtitle=xtitles[1], ytitle=ytitles[1], xum=xums[1], yum=yums[1])
                        self._plot_xy_detailed(output_beam, progressBarValue + 12, variables[2][0], variables[2][1], plot_canvas_index=2, title=titles[2], xtitle=xtitles[2], ytitle=ytitles[2], xum=xums[2], yum=yums[2])
                        self._plot_xy_detailed(output_beam, progressBarValue + 16, variables[3][0], variables[3][1], plot_canvas_index=3, title=titles[3], xtitle=xtitles[3], ytitle=ytitles[3], xum=xums[3], yum=yums[3])
                        self._plot_histo_detailed(output_beam, progressBarValue + 20, variables[4], plot_canvas_index=4, title=titles[4], xtitle=xtitles[4], ytitle=ytitles[4], xum=xums[4])
                        if self.has_footprint: self._plot_xy_detailed(footprint, progressBarValue + 20, 2, 1, plot_canvas_index=5, title="Footprint", xtitle="Y [m]", ytitle="X [m]", xum=("Y [m]"), yum=("X [m]"), is_footprint=True)

                except Exception as e:
                    self.view_type_combo.setEnabled(True)

                    raise Exception("Data not plottable: No good rays or bad content\nexception: " + str(e))

                self.view_type_combo.setEnabled(True)
            else:
                raise Exception("Empty Beam")

        self.plotted_beam   = output_beam
        self.footprint_beam = footprint

    def _write_stdout(self, text):
        cursor = self.shadow_output.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(text)
        self.shadow_output.setTextCursor(cursor)
        self.shadow_output.ensureCursorVisible()

    def _on_receiving_input(self):
        self._initialize_tabs()

    def _get_variables_to_plot(self):
        return [[1, 3], [4, 6], [1, 4], [3, 6], 26]

    def _get_titles(self):
        return ["X,Z", "X',Z'", "X,X'", "Z,Z'", "Energy"]

    def _get_x_titles(self):
        return [r'X [$\mu$m]', "X' [$\mu$rad]", r'X [$\mu$m]', r'Z [$\mu$m]', "Energy [eV]"]

    def _get_y_titles(self):
        return [r'Z [$\mu$m]', "Z' [$\mu$rad]", "X' [$\mu$rad]", "Z' [$\mu$rad]", "Number of Rays"]

    def _get_x_um(self):
        return ["X [" + u"\u03BC" + "m]", "X' [" + u"\u03BC" + "rad]", "X [" + u"\u03BC" + "m]", "Z [" + u"\u03BC" + "m]", "[eV]"]

    def _get_y_um(self):
        return ["Z [" + u"\u03BC" + "m]", "Z' [" + u"\u03BC" + "rad]", "X' [" + u"\u03BC" + "rad]", "Z' [" + u"\u03BC" + "rad]", None]

    def _is_conversion_active(self):
        return True

if __name__ == "__main__":
    a = QApplication(sys.argv)
    ow = GenericElement()
    ow.show()
    a.exec_()
    ow.saveSettings()



from orangewidget.settings import Setting
from oasys.widgets import gui as oasysgui
from syned.beamline.shape import Circle
from shadow4.beamline.optical_elements.refractors.s4_crl import S4CRL, S4CRLElement
from orangecontrib.shadow4.widgets.gui.ow_abstract_lens import OWAbstractLens

class OWCRL(OWAbstractLens):
    name = "Compound Refractive Lens"
    description = "Shadow Compound Refractive Lens"
    icon = "icons/crl.png"
    priority = 2.2

    n_lens                  = Setting(10)
    piling_thickness        = Setting(2.5)

    def __init__(self):
        super().__init__()

    def create_basic_settings_subtabs(self, tabs_basic_settings):
        return oasysgui.createTabPage(tabs_basic_settings, "CRL") # to be populated

    def populate_basic_setting_subtabs(self, basic_setting_subtabs):
        crl_box = oasysgui.widgetBox(basic_setting_subtabs, "CRL Parameters", addSpace=False, orientation="vertical", height=90)

        oasysgui.lineEdit(crl_box, self, "n_lens", "Number of lenses", labelWidth=260, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(crl_box, self, "piling_thickness", "Piling thickness [mm]", labelWidth=260, valueType=float, orientation="horizontal")

        super(OWCRL, self).populate_basic_setting_subtabs(basic_setting_subtabs)

    def get_optical_element_instance(self):
        try:    name = self.getNode().title
        except: name = "Compound Refractive Lens"

        um_to_si = 1e-6
        mm_to_si = 1e-3

        if self.has_finite_diameter == 0: boundary_shape = Circle(radius=um_to_si * self.diameter * 0.5)
        else:                             boundary_shape = None
        if self.is_cylinder == 1: cylinder_angle = self.cylinder_angle + 1
        else:                     cylinder_angle = 0

        return S4CRL(name=name,
                     n_lens=self.n_lens,
                     piling_thickness=self.piling_thickness*mm_to_si,
                     boundary_shape=boundary_shape,
                     material="", # not used
                     thickness=self.interthickness * um_to_si,
                     surface_shape=self.surface_shape,
                     convex_to_the_beam=self.convex_to_the_beam,
                     cylinder_angle=cylinder_angle,
                     ri_calculation_mode=self.ri_calculation_mode,
                     prerefl_file=self.prerefl_file,
                     refraction_index=self.refraction_index,
                     attenuation_coefficient=self.attenuation_coefficient,
                     radius=self.radius * um_to_si,
                     conic_coefficients=None)  # TODO: add conic coefficient shape to the GUI

    def get_beamline_element_instance(self):
        return S4CRLElement()



import sys
import time
import copy
import numpy

from PyQt5.QtGui import QTextCursor
from PyQt5.QtCore import QSettings
from orangewidget import gui
from orangewidget.settings import Setting
from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence
from oasys.widgets.gui import ConfirmDialog, MessageDialog, selectSaveFileFromDialog
from oasys.util.oasys_util import EmittingStream

from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence, ShadowPlot
from orangecontrib.shadow4.widgets.gui.ow_automatic_element import AutomaticElement
from shadow4.beam.s4_beam import S4Beam

class Histogram(AutomaticElement):

    name = "Histogram"
    description = "Display Data Tools: Histogram"
    icon = "icons/histogram.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 1.0
    category = "Display Data Tools"
    keywords = ["data", "file", "load", "read"]

    inputs = [("Shadow Data", ShadowData, "set_shadow_data")]

    IMAGE_WIDTH  = 878
    IMAGE_HEIGHT = 635

    want_main_area = 1

    plot_canvas = None
    input_data  = None

    image_plane                  = Setting(0)
    image_plane_new_position     = Setting(10.0)
    image_plane_rel_abs_position = Setting(0)

    x_column_index = Setting(25)

    x_range     = Setting(0)
    x_range_min = Setting(0.0)
    x_range_max = Setting(0.0)

    weight_column_index = Setting(22)
    rays                 = Setting(1)

    number_of_bins = Setting(100)

    title = Setting("Energy")

    autosave           = Setting(0)
    autosave_file_name = Setting("autosave_histogram_plot.hdf5")

    keep_result              = Setting(0)
    autosave_partial_results = Setting(0)

    conversion_active = Setting(1)

    cumulated_ticket = None
    plotted_ticket   = None
    autosave_file    = None
    autosave_prog_id = 0

    def __init__(self):
        super().__init__()

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        gui.button(button_box, self, "Refresh", callback=self.plot_results, height=45)
        gui.button(button_box, self, "Save Current Plot", callback=self.save_results, height=45)

        gui.separator(self.controlArea, 10)

        self.tabs_setting = oasysgui.tabWidget(self.controlArea)
        self.tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        # graph tab
        tab_set = oasysgui.createTabPage(self.tabs_setting, "Plot Settings")
        tab_gen = oasysgui.createTabPage(self.tabs_setting, "Histogram Settings")

        screen_box = oasysgui.widgetBox(tab_set, "Screen Position Settings", addSpace=True, orientation="vertical", height=120)

        self.image_plane_combo = gui.comboBox(screen_box, self, "image_plane", label="Position of the Image",
                                              items=["On Image Plane", "Retraced"], labelWidth=260,
                                              callback=self.set_image_plane, sendSelectedValue=False, orientation="horizontal")

        self.image_plane_box = oasysgui.widgetBox(screen_box, "", addSpace=False, orientation="vertical", height=50)
        self.image_plane_box_empty = oasysgui.widgetBox(screen_box, "", addSpace=False, orientation="vertical", height=50)

        oasysgui.lineEdit(self.image_plane_box, self, "image_plane_new_position", "Image Plane new Position", labelWidth=220, valueType=float, orientation="horizontal")

        gui.comboBox(self.image_plane_box, self, "image_plane_rel_abs_position", label="Position Type", labelWidth=250,
                     items=["Absolute", "Relative"], sendSelectedValue=False, orientation="horizontal")

        self.set_image_plane()

        general_box = oasysgui.widgetBox(tab_set, "General Settings", addSpace=True, orientation="vertical", height=250)

        self.x_column = gui.comboBox(general_box, self, "x_column_index", label="Column", labelWidth=70,
                                     items=S4Beam.column_names_with_column_number(),
                                     sendSelectedValue=False, orientation="horizontal", callback=self.set_column_index)

        gui.comboBox(general_box, self, "x_range", label="Range", labelWidth=250,
                     items=["<Default>", "Set.."],
                     callback=self.set_x_range, sendSelectedValue=False, orientation="horizontal")

        self.x_range_box = oasysgui.widgetBox(general_box, "", addSpace=True, orientation="vertical", height=100)
        self.x_range_box_empty = oasysgui.widgetBox(general_box, "", addSpace=True, orientation="vertical", height=100)

        self.le_x_range_min = oasysgui.lineEdit(self.x_range_box, self, "x_range_min", "Min", labelWidth=220, valueType=float, orientation="horizontal")
        self.le_x_range_max = oasysgui.lineEdit(self.x_range_box, self, "x_range_max", "Max", labelWidth=220, valueType=float, orientation="horizontal")

        self.set_x_range()

        col_names = S4Beam.column_names_with_column_number()
        col_names.insert(0, "0: No Weight")
        print(col_names)
        self.weight_column = gui.comboBox(general_box, self, "weight_column_index", label="Weight", labelWidth=70,
                                         items=col_names,
                                         sendSelectedValue=False, orientation="horizontal")

        gui.comboBox(general_box, self, "rays", label="Rays", labelWidth=250,
                                     items=["All rays",
                                            "Good Only",
                                            "Lost Only"],
                                     sendSelectedValue=False, orientation="horizontal")

        autosave_box = oasysgui.widgetBox(tab_gen, "Autosave", addSpace=True, orientation="vertical", height=85)

        gui.comboBox(autosave_box, self, "autosave", label="Save automatically plot into file", labelWidth=250,
                                         items=["No", "Yes"],
                                         sendSelectedValue=False, orientation="horizontal", callback=self.set_autosave)

        self.autosave_box_1 = oasysgui.widgetBox(autosave_box, "", addSpace=False, orientation="horizontal", height=25)
        self.autosave_box_2 = oasysgui.widgetBox(autosave_box, "", addSpace=False, orientation="horizontal", height=25)

        self.le_autosave_file_name = oasysgui.lineEdit(self.autosave_box_1, self, "autosave_file_name", "File Name", labelWidth=100,  valueType=str, orientation="horizontal")

        gui.button(self.autosave_box_1, self, "...", callback=self.select_autosave_file)

        incremental_box = oasysgui.widgetBox(tab_gen, "Incremental Result", addSpace=True, orientation="vertical", height=120)

        gui.comboBox(incremental_box, self, "keep_result", label="Keep Result", labelWidth=250,
                     items=["No", "Yes"], sendSelectedValue=False, orientation="horizontal", callback=self.set_autosave)

        self.cb_autosave_partial_results = gui.comboBox(incremental_box, self, "autosave_partial_results", label="Save partial plots into file", labelWidth=250,
                                                        items=["No", "Yes"], sendSelectedValue=False, orientation="horizontal")

        gui.button(incremental_box, self, "Clear", callback=self.clear_results)

        histograms_box = oasysgui.widgetBox(tab_gen, "Histograms settings", addSpace=True, orientation="vertical", height=90)

        oasysgui.lineEdit(histograms_box, self, "number_of_bins", "Number of Bins", labelWidth=250, valueType=int, orientation="horizontal")

        gui.comboBox(histograms_box, self, "conversion_active", label="Is U.M. conversion active", labelWidth=250,
                     items=["No", "Yes"], sendSelectedValue=False, orientation="horizontal", callback=self.set_is_conversion_active)

        self.set_autosave()

        self.main_tabs = oasysgui.tabWidget(self.mainArea)
        plot_tab = oasysgui.createTabPage(self.main_tabs, "Plots")
        out_tab = oasysgui.createTabPage(self.main_tabs, "Output")

        self.image_box = gui.widgetBox(plot_tab, "Plot Result", addSpace=True, orientation="vertical")
        self.image_box.setFixedHeight(self.IMAGE_HEIGHT)
        self.image_box.setFixedWidth(self.IMAGE_WIDTH)

        self.shadow_output = oasysgui.textArea(height=580, width=800)

        out_box = gui.widgetBox(out_tab, "System Output", addSpace=True, orientation="horizontal")
        out_box.layout().addWidget(self.shadow_output)

    def clear_results(self, interactive=True):
        if not interactive: proceed = True
        else: proceed = ConfirmDialog.confirmed(parent=self)

        if proceed:
            self.input_data = ShadowData()
            self.cumulated_ticket = None
            self.plotted_ticket = None
            self.autosave_prog_id = 0
            if not self.autosave_file is None:
                self.autosave_file.close()
                self.autosave_file = None

            self.plot_canvas.clear()

    def set_column_index(self):
        self.__change_labels()

    def set_is_conversion_active(self):
        self.__change_labels()

    def set_x_range(self):
        self.x_range_box.setVisible(self.x_range == 1)
        self.x_range_box_empty.setVisible(self.x_range == 0)
        self.__change_labels()

    def __change_labels(self):
        def change_label(line_edit, index):
            label      = line_edit.parent().layout().itemAt(0).widget()
            label_text = label.text()
            if label_text[-1] == "]": label_text = label_text.split(sep="[")[0]
            else: label_text += " "
            if    index in [0, 1, 2] and self.is_conversion_active(): label_text += "[\u03BCm]"
            elif  index in [3, 4, 5] and self.is_conversion_active(): label_text += "[\u03BCrad]"
            else:                                                      label_text += S4Beam.column_units()[index]
            label.setText(label_text)

        if self.x_range == 1:
            change_label(self.le_x_range_min, self.x_column_index)
            change_label(self.le_x_range_max, self.x_column_index)

    def set_image_plane(self):
        self.image_plane_box.setVisible(self.image_plane==1)
        self.image_plane_box_empty.setVisible(self.image_plane==0)

    def set_autosave(self):
        self.autosave_box_1.setVisible(self.autosave==1)
        self.autosave_box_2.setVisible(self.autosave==0)

        self.cb_autosave_partial_results.setEnabled(self.autosave==1 and self.keep_result==1)

    def select_autosave_file(self):
        self.le_autosave_file_name.setText(oasysgui.selectFileFromDialog(self, self.autosave_file_name, "Select File", file_extension_filter="HDF5 Files (*.hdf5 *.h5 *.hdf)"))

    def replace_histo(self, beam, var, x_range, title, xtitle, ytitle, xum, flux):
        if self.plot_canvas is None:
            self.plot_canvas = ShadowPlot.DetailedHistoWidget(y_scale_factor=1.14)
            self.image_box.layout().addWidget(self.plot_canvas)

        try:
            if self.autosave == 1:
                if self.autosave_file is None:
                    self.autosave_file = ShadowPlot.HistogramHdf5File(congruence.checkDir(self.autosave_file_name))
                elif self.autosave_file.filename != congruence.checkFileName(self.autosave_file_name):
                    self.autosave_file.close()
                    self.autosave_file = ShadowPlot.HistogramHdf5File(congruence.checkDir(self.autosave_file_name))

            if self.keep_result == 1:
                self.cumulated_ticket, last_ticket = self.plot_canvas.plot_histo(beam, var, self.rays, x_range, self.weight_column_index, title, xtitle, ytitle,
                                                                                 nbins=self.number_of_bins,
                                                                                 xum=xum,
                                                                                 ticket_to_add=self.cumulated_ticket,
                                                                                 flux=flux)

                self.plotted_ticket = self.cumulated_ticket

                if self.autosave == 1:
                    self.autosave_prog_id += 1
                    self.autosave_file.write_coordinates(self.cumulated_ticket)
                    dataset_name = self.weight_column.itemText(self.weight_column_index)

                    self.autosave_file.add_histogram(self.cumulated_ticket, dataset_name=dataset_name)

                    if self.autosave_partial_results == 1:
                        if last_ticket is None: self.autosave_file.add_histogram(self.cumulated_ticket, plot_name="Histogram #" + str(self.autosave_prog_id), dataset_name=dataset_name)
                        else:                   self.autosave_file.add_histogram(last_ticket, plot_name="Histogram #" + str(self.autosave_prog_id), dataset_name=dataset_name)

                    self.autosave_file.flush()
            else:
                ticket, _ = self.plot_canvas.plot_histo(beam, var, self.rays, x_range, self.weight_column_index, title, xtitle, ytitle,
                                                        nbins=self.number_of_bins,
                                                        xum=xum,
                                                        flux=flux)

                self.cumulated_ticket = None
                self.plotted_ticket = ticket

                if self.autosave == 1:
                    self.autosave_prog_id += 1
                    self.autosave_file.write_coordinates(ticket)
                    self.autosave_file.add_histogram(ticket, dataset_name=self.weight_column.itemText(self.weight_column_index))
                    self.autosave_file.flush()

        except Exception as e:
            if not self.IS_DEVELOP:
                raise Exception("Data not plottable: Bad content")
            else:
                raise e

    def plot_histo(self, var_x, title, xtitle, ytitle, xum):
        beam_to_plot = self.input_data.beam
        flux         = self.input_data.get_flux(nolost=self.rays)

        if self.image_plane == 1:
            new_shadow_beam = self.input_data.beam.duplicate()
            dist = 0.0

            if self.image_plane_rel_abs_position == 1:  # relative
                dist = self.image_plane_new_position
            else:  # absolute
                if self.input_data.beamline is None == 0: beamline_element = None
                else:                                     beamline_element = self.input_data.beamline.get_beamline_element_at(-1)

                if beamline_element is None: image_plane = 0.0
                else:                        image_plane = beamline_element.get_coordinates().q()

                dist = self.image_plane_new_position - image_plane

            self.retrace_beam(new_shadow_beam, dist)

            beam_to_plot = new_shadow_beam

        x_range = self.get_range(beam_to_plot, var_x)

        self.replace_histo(beam_to_plot, var_x, x_range, title, xtitle, ytitle, xum, flux)

    def get_range(self, beam_to_plot : S4Beam, var_x):
        if self.x_range == 0 :
            x_max = 0
            x_min = 0

            x, good_only = beam_to_plot.get_columns((var_x, 10))

            x_to_plot = copy.deepcopy(x)

            go = numpy.where(good_only == 1)
            lo = numpy.where(good_only != 1)

            if self.rays == 0:
                x_max = numpy.array(x_to_plot[0:], float).max()
                x_min = numpy.array(x_to_plot[0:], float).min()
            elif self.rays == 1:
                x_max = numpy.array(x_to_plot[go], float).max()
                x_min = numpy.array(x_to_plot[go], float).min()
            elif self.rays == 2:
                x_max = numpy.array(x_to_plot[lo], float).max()
                x_min = numpy.array(x_to_plot[lo], float).min()

            x_range = [x_min, x_max]
        else:
            congruence.checkLessThan(self.x_range_min, self.x_range_max, "X range min", "X range max")

            factor = ShadowPlot.get_factor(var_x)
            x_range = [self.x_range_min / factor, self.x_range_max / factor]

        return x_range

    def save_results(self):
        if not self.plotted_ticket is None:
            try:
                file_name = oasysgui.selectSaveFileFromDialog(self, message="Save Current Plot", file_extension_filter="HDF5 Files (*.hdf5 *.h5 *.hdf)")

                if not file_name is None and not file_name.strip() == "":
                    if not (file_name.endswith("hd5") or file_name.endswith("hdf5") or file_name.endswith("hdf")): file_name += ".hdf5"

                    save_file = ShadowPlot.HistogramHdf5File(congruence.checkDir(file_name))
                    save_file.write_coordinates(self.plotted_ticket)
                    save_file.add_histogram(self.plotted_ticket, dataset_name=self.weight_column.itemText(self.weight_column_index))

                    save_file.close()
            except Exception as exception:
                self.prompt_exception(exception)

    def plot_results(self):
        try:
            plotted = False

            sys.stdout = EmittingStream(textWritten=self.write_stdout)

            if ShadowCongruence.check_empty_data(self.input_data):
                ShadowPlot.set_conversion_active(self.is_conversion_active())

                self.number_of_bins = congruence.checkPositiveNumber(self.number_of_bins, "Number of Bins")

                x, auto_title, xum = self.get_titles()

                self.plot_histo(x, title=self.title, xtitle=auto_title, ytitle="Number of Rays", xum=xum)

                plotted = True

            time.sleep(0.1)  # prevents a misterious dead lock in the Orange cycle when refreshing the histogram

            return plotted
        except Exception as exception:
            self.prompt_exception(exception)

    def get_titles(self):
        self.title = xum = auto_title = self.x_column.currentText()
        x = self.x_column_index + 1

        if x in [1, 2, 3] and self.is_conversion_active():
            xum        += " [\u03BCm]"
            auto_title += " [$\mu$m]"
        elif x in [4, 5, 6] and self.is_conversion_active():
            xum        += " [\u03BCrad]"
            auto_title += " [$\mu$rad]"
        else:
            xum        += " " + S4Beam.column_units()[self.x_column_index]
            auto_title += " " + S4Beam.column_units()[self.x_column_index]

        return x, auto_title, xum

    def set_shadow_data(self, shadow_data : ShadowData):
        if ShadowCongruence.check_empty_data(shadow_data):
            if ShadowCongruence.check_empty_beam(shadow_data.beam):
                self.input_data = shadow_data
                if self.is_automatic_run: self.plot_results()
            else:
                MessageDialog.message(self, "Data not displayable: bad content", "Error", "critical")

    def write_stdout(self, text):
        cursor = self.shadow_output.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(text)
        self.shadow_output.setTextCursor(cursor)
        self.shadow_output.ensureCursorVisible()

    def retrace_beam(self, new_shadow_beam : S4Beam, dist):
        new_shadow_beam.retrace(dist)

    def is_conversion_active(self):
        return self.conversion_active == 1



import numpy
import sys

from PyQt5.QtGui import QPalette, QColor, QFont

from orangewidget import widget
from orangewidget import gui
from orangewidget.settings import Setting

from oasys.widgets import gui as oasysgui
from oasys.util.oasys_util import EmittingStream

from syned.widget.widget_decorator import WidgetDecorator
from syned.beamline.element_coordinates import ElementCoordinates

from orangecontrib.shadow4.widgets.gui.ow_generic_element import GenericElement
from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence

NO_FILE_SPECIFIED = "<specify file name>"
SUBTAB_INNER_BOX_WIDTH = 375


def optical_element_inputs():
    inputs = [("Shadow Data", ShadowData, "set_shadow_data")]
    WidgetDecorator.append_syned_input_data(inputs)

    return inputs

class OWOpticalElement(GenericElement, WidgetDecorator):
    inputs  = optical_element_inputs()
    outputs = [{"name":"Shadow Data", "type":ShadowData, "doc":"",}]

    #########################################################
    # Position
    #########################################################
    source_plane_distance           = Setting(1.0)
    image_plane_distance            = Setting(1.0)
    angles_respect_to               = Setting(0)
    incidence_angle_deg             = Setting(88.8)
    incidence_angle_mrad            = Setting(0.0)
    reflection_angle_deg            = Setting(85.0)
    reflection_angle_mrad           = Setting(0.0)
    oe_orientation_angle            = Setting(0)
    oe_orientation_angle_user_value = Setting(0.0)

    def __init__(self, show_automatic_box=True, has_footprint=False):
        super().__init__(show_automatic_box=show_automatic_box, has_footprint=has_footprint)

        #
        # main buttons
        #
        self.runaction = widget.OWAction("Run Shadow4/Trace", self)
        self.runaction.triggered.connect(self.run_shadow4)
        self.addAction(self.runaction)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Run shadow4/trace", callback=self.run_shadow4)
        font = QFont(button.font())
        font.setBold(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Blue'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)

        button = gui.button(button_box, self, "Reset Fields", callback=self.call_reset_settings)
        font = QFont(button.font())
        font.setItalic(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Red'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)
        button.setFixedWidth(150)

        #
        # tabs
        #
        self.tabs_control_area = oasysgui.tabWidget(self.controlArea)
        self.tabs_control_area.setFixedHeight(self.TABS_AREA_HEIGHT)
        self.tabs_control_area.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        self.tab_position          = oasysgui.createTabPage(self.tabs_control_area, "Position")           # to be populated
        self.tab_basic_settings    = oasysgui.createTabPage(self.tabs_control_area, "Basic Settings")
        self.tab_advanced_settings = oasysgui.createTabPage(self.tabs_control_area, "Advanced Settings")

        self.tabs_basic_settings   = oasysgui.tabWidget(self.tab_basic_settings)
        basic_setting_subtabs      = self.create_basic_settings_subtabs(self.tabs_basic_settings)

        self.tabs_advanced_settings   = oasysgui.tabWidget(self.tab_advanced_settings)

        advanced_setting_subtabs = self.create_advanced_settings_subtabs(self.tabs_advanced_settings)

        #########################################################
        # Position
        #########################################################
        self.populate_tab_position(self.tab_position)

        #########################################################
        # Basic Settings
        #########################################################

        self.populate_basic_setting_subtabs(basic_setting_subtabs)

        #########################################################
        # Advanced Settings
        #########################################################
        self.populate_advanced_setting_subtabs(advanced_setting_subtabs)

        gui.rubber(self.controlArea)
        gui.rubber(self.mainArea)

    def create_basic_settings_subtabs(self, tabs_basic_settings): return None
    def create_advanced_settings_subtabs(self, tabs_advanced_settings): return None

    def populate_basic_setting_subtabs(self, basic_setting_subtabs): pass
    def populate_advanced_setting_subtabs(self, advanced_setting_subtabs): pass

    def populate_tab_position(self, tab_position):
        self.orientation_box = oasysgui.widgetBox(tab_position, "Optical Element Orientation", addSpace=True, orientation="vertical")

        oasysgui.lineEdit(self.orientation_box, self, "source_plane_distance", "Source Plane Distance [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="source_plane_distance")
        oasysgui.lineEdit(self.orientation_box, self, "image_plane_distance", "Image Plane Distance [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="image_plane_distance")

        gui.comboBox(self.orientation_box, self, "angles_respect_to", label="Angles in [deg] with respect to the",
                     labelWidth=250, items=["Normal", "Surface"], callback=self.set_angles_respect_to,
                     sendSelectedValue=False, orientation="horizontal", tooltip="angles_respect_to")

        self.incidence_angle_deg_le = oasysgui.lineEdit(self.orientation_box, self, "incidence_angle_deg",
                                                        "Incident Angle\nwith respect to the Normal [deg]",
                                                        labelWidth=220, callback=self.calculate_incidence_angle_mrad,
                                                        valueType=float, orientation="horizontal", tooltip="incidence_angle_deg")
        self.incidence_angle_rad_le = oasysgui.lineEdit(self.orientation_box, self, "incidence_angle_mrad",
                                                        "Incident Angle\nwith respect to the surface [mrad]",
                                                        labelWidth=220, callback=self.calculate_incidence_angle_deg,
                                                        valueType=float, orientation="horizontal", tooltip="incidence_angle_mrad")
        self.reflection_angle_deg_le = oasysgui.lineEdit(self.orientation_box, self, "reflection_angle_deg",
                                                         "Reflection Angle\nwith respect to the Normal [deg]",
                                                         labelWidth=220, callback=self.calculate_reflection_angle_mrad,
                                                         valueType=float, orientation="horizontal", tooltip="reflection_angle_deg")
        self.reflection_angle_rad_le = oasysgui.lineEdit(self.orientation_box, self, "reflection_angle_mrad",
                                                         "Reflection Angle\nwith respect to the surface [mrad]",
                                                         labelWidth=220, callback=self.calculate_reflection_angle_deg,
                                                         valueType=float, orientation="horizontal", tooltip="reflection_angle_mrad")

        self.set_angles_respect_to()

        self.calculate_incidence_angle_mrad()
        self.calculate_reflection_angle_mrad()

        gui.comboBox(self.orientation_box, self, "oe_orientation_angle", label="O.E. Orientation Angle [deg]",
                     labelWidth=390,
                     items=[0, 90, 180, 270, "Other value..."],
                     valueType=float,
                     sendSelectedValue=False, orientation="horizontal", callback=self.oe_orientation_angle_user,
                     tooltip="oe_orientation_angle" )
        self.oe_orientation_angle_user_value_le = oasysgui.widgetBox(self.orientation_box, "", addSpace=False,
                                                                         orientation="vertical")
        oasysgui.lineEdit(self.oe_orientation_angle_user_value_le, self, "oe_orientation_angle_user_value",
                          "O.E. Orientation Angle [deg]",
                          labelWidth=220,
                          valueType=float, orientation="horizontal", tooltip="oe_orientation_angle_user_value")

        self.oe_orientation_angle_user()


    #########################################################
    # Position Methods
    #########################################################
    def set_angles_respect_to(self):
        label_1 = self.incidence_angle_deg_le.parent().layout().itemAt(0).widget()
        label_2 = self.reflection_angle_deg_le.parent().layout().itemAt(0).widget()

        if self.angles_respect_to == 0:
            label_1.setText("Incident Angle\nwith respect to the normal [deg]")
            label_2.setText("Reflection Angle\nwith respect to the normal [deg]")
        else:
            label_1.setText("Incident Angle\nwith respect to the surface [deg]")
            label_2.setText("Reflection Angle\nwith respect to the surface [deg]")

        self.calculate_incidence_angle_mrad()
        self.calculate_reflection_angle_mrad()

    def calculate_incidence_angle_mrad(self):
        digits = 7

        if self.angles_respect_to == 0: self.incidence_angle_mrad = round(numpy.radians(90-self.incidence_angle_deg)*1000, digits)
        else:                           self.incidence_angle_mrad = round(numpy.radians(self.incidence_angle_deg)*1000, digits)

    def calculate_reflection_angle_mrad(self):
        digits = 7
        if self.angles_respect_to == 0: self.reflection_angle_mrad = round(numpy.radians(90 - self.reflection_angle_deg)*1000, digits)
        else:                           self.reflection_angle_mrad = round(numpy.radians(self.reflection_angle_deg)*1000, digits)

    def calculate_incidence_angle_deg(self):
        digits = 10
        if self.angles_respect_to == 0: self.incidence_angle_deg = round(numpy.degrees(0.5 * numpy.pi - (self.incidence_angle_mrad / 1000)), digits)
        else:                           self.incidence_angle_deg = round(numpy.degrees(self.incidence_angle_mrad / 1000), digits)

    def calculate_reflection_angle_deg(self):
        digits = 10

        if self.angles_respect_to == 0: self.reflection_angle_deg = round(numpy.degrees(0.5*numpy.pi-(self.reflection_angle_mrad/1000)), digits)
        else:                           self.reflection_angle_deg = round(numpy.degrees(self.reflection_angle_mrad/1000), digits)

    def oe_orientation_angle_user(self):
        if self.oe_orientation_angle < 4: self.oe_orientation_angle_user_value_le.setVisible(False)
        else:                             self.oe_orientation_angle_user_value_le.setVisible(True)

    def get_oe_orientation_angle(self):
        if self.oe_orientation_angle == 0:   return 0.0
        elif self.oe_orientation_angle == 1: return 90.0
        elif self.oe_orientation_angle == 2: return 180.0
        elif self.oe_orientation_angle == 3: return 270.0
        elif self.oe_orientation_angle == 4: return self.oe_orientation_angle_user_value

    def get_coordinates_instance(self):
        if self.angles_respect_to == 0:
            angle_radial = numpy.radians(self.incidence_angle_deg)
            angle_radial_out = numpy.radians(self.reflection_angle_deg)
        elif self.angles_respect_to == 1:
            angle_radial = numpy.pi / 2 - self.incidence_angle_mrad * 1e-3
            angle_radial_out = numpy.pi / 2 - self.reflection_angle_mrad * 1e-3

        # angle_radial = numpy.pi / 2 - self.incidence_angle_mrad * 1e-3
        # angle_radial_out = numpy.pi / 2 - self.reflection_angle_mrad * 1e-3

        print(">>>>>>normal inc ref [deg]:", numpy.degrees(angle_radial), numpy.degrees(angle_radial_out), self.get_oe_orientation_angle())
        print(">>>>>>grazing inc ref [mrad]:", 1e3 * (numpy.pi / 2 - angle_radial), 1e3 * (numpy.pi / 2 - angle_radial_out))
        print(">>>>>>m.o.a. [deg]:", self.get_oe_orientation_angle())

        return ElementCoordinates(
                p=self.source_plane_distance,
                q=self.image_plane_distance,
                angle_radial=angle_radial,
                angle_azimuthal=numpy.radians(self.get_oe_orientation_angle()),
                angle_radial_out=angle_radial_out,
                )


    def set_shadow_data(self, input_data):
        self.not_interactive = self._check_not_interactive_conditions(input_data)

        self._on_receiving_input()

        if ShadowCongruence.check_empty_data(input_data):
            self.input_data = input_data.duplicate()
            if self.is_automatic_run: self.run_shadow4()

    def run_shadow4(self):
        self.shadow_output.setText("")

        sys.stdout = EmittingStream(textWritten=self._write_stdout)

        try:

            beamline = self.input_data.beamline.duplicate()
            element = self.get_beamline_element_instance()
            element.set_optical_element(self.get_optical_element_instance())
            element.set_coordinates(self.get_coordinates_instance())
            element.set_movements(self.get_movements_instance())
            element.set_input_beam(self.input_data.beam)

            print(element.info())

            beamline.append_beamline_element(element)

            #
            # script
            #
            script = beamline.to_python_code()
            script += "\n\n\n# test plot"
            script += "\nif True:"
            script += "\n   from srxraylib.plot.gol import plot_scatter"
            script += "\n   plot_scatter(beam.get_photon_energy_eV(nolost=1), beam.get_column(23, nolost=1), title='(Intensity,Photon Energy)', plot_histograms=0)"
            script += "\n   plot_scatter(1e6 * beam.get_column(1, nolost=1), 1e6 * beam.get_column(3, nolost=1), title='(X,Z) in microns')"
            self.shadow4_script.set_code(script)

            #
            # run
            #
            self.progressBarInit()
            output_beam, footprint = element.trace_beam()
            self._set_plot_quality()
            self._plot_results(output_beam, footprint, progressBarValue=80)
            self.progressBarFinished()

            #
            # send beam
            #
            self.send("Shadow Data", ShadowData(beam=output_beam, beamline=beamline))
        except Exception as exception:
            self.prompt_exception(exception)
            self._initialize_tabs()

    def receive_syned_data(self, data): raise Exception("Not yet implemented")

    def get_optical_element_instance(self): raise NotImplementedError()
    def get_beamline_element_instance(self): raise NotImplementedError()
    def get_movements_instance(self): return None




from orangewidget import gui
from orangewidget.settings import Setting

from oasys.widgets import gui as oasysgui

from shadow4.beamline.optical_elements.refractors.s4_conic_interface import S4ConicInterface, S4ConicInterfaceElement

from orangecontrib.shadow4.widgets.gui.ow_optical_element_with_surface_shape import OWOpticalElementWithSurfaceShape

class OWRefractiveInterface(OWOpticalElementWithSurfaceShape):
    name        = "Refractive Interface"
    description = "Shadow Refractive Interface"
    icon        = "icons/interface.png"

    optical_constants_refraction_index = Setting(0)
    refraction_index_in_object_medium  = Setting(0.0)
    attenuation_in_object_medium       = Setting(0.0)
    file_prerefl_for_object_medium     = Setting("<none>")
    refractive_index_in_image_medium   = Setting(0.0)
    attenuation_in_image_medium        = Setting(0.0)
    file_prerefl_for_image_medium      = Setting("<none>")

    priority = 3.1

    def __init__(self):
        super().__init__(has_footprint=False, switch_icons=False)

    def create_basic_settings_specific_subtabs(self, tabs_basic_settings):
        return oasysgui.createTabPage(tabs_basic_settings, "Refractor")

    def populate_basic_settings_specific_subtabs(self, specific_basic_settings_subtabs):
        tab_bas_refractor = specific_basic_settings_subtabs

        refractor_box = oasysgui.widgetBox(tab_bas_refractor, "Optical Constants - Refractive Index", addSpace=False, orientation="vertical", height=320)

        gui.comboBox(refractor_box, self, "optical_constants_refraction_index", label="optical constants\n/refraction index", labelWidth=120,
                     items=["from constants in both media",
                            "from prerefl in OBJECT media",
                            "from prerefl in IMAGE media",
                            "from prerefl in both media"],
                     callback=self.set_refractor_optical_constants, sendSelectedValue=False, orientation="horizontal")

        gui.separator(refractor_box, height=10)
        self.refractor_object_box_1 = oasysgui.widgetBox(refractor_box, "OBJECT side", addSpace=False, orientation="vertical", height=100)
        oasysgui.lineEdit(self.refractor_object_box_1, self, "refraction_index_in_object_medium", "refraction index in object medium", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_attenuation_in_object_medium = oasysgui.lineEdit(self.refractor_object_box_1, self, "attenuation_in_object_medium", "attenuation in object medium", labelWidth=260, valueType=float, orientation="horizontal")

        self.refractor_object_box_2 = oasysgui.widgetBox(refractor_box, "OBJECT side", addSpace=False, orientation="horizontal", height=100)
        self.le_file_prerefl_for_object_medium = oasysgui.lineEdit(self.refractor_object_box_2, self, "file_prerefl_for_object_medium",
                                                                   "file prerefl for\nobject medium", labelWidth=120, valueType=str, orientation="horizontal")

        gui.button(self.refractor_object_box_2, self, "...", callback=self.select_prerefl_object_file_name)

        self.refractor_image_box_1 = oasysgui.widgetBox(refractor_box, "IMAGE side", addSpace=False, orientation="vertical", height=100)
        oasysgui.lineEdit(self.refractor_image_box_1, self, "refractive_index_in_image_medium", "refraction index in image medium", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_attenuation_in_image_medium = oasysgui.lineEdit(self.refractor_image_box_1, self, "attenuation_in_image_medium", "attenuation in image medium", labelWidth=260, valueType=float, orientation="horizontal")

        self.refractor_image_box_2 = oasysgui.widgetBox(refractor_box, "IMAGE side", addSpace=False, orientation="horizontal", height=100)
        self.le_file_prerefl_for_image_medium = oasysgui.lineEdit(self.refractor_image_box_2, self, "file_prerefl_for_image_medium",
                                                                  "file prerefl for\nimage medium", labelWidth=120, valueType=str, orientation="horizontal")

        gui.button(self.refractor_image_box_2, self, "...", callback=self.select_prerefl_image_file_name)

        self.set_refractor_optical_constants()

    def set_refractor_optical_constants(self):
        self.refractor_object_box_1.setVisible(self.optical_constants_refraction_index == 0 or self.optical_constants_refraction_index == 2)
        self.refractor_object_box_2.setVisible(self.optical_constants_refraction_index == 1 or self.optical_constants_refraction_index == 3)
        self.refractor_image_box_1.setVisible(self.optical_constants_refraction_index == 0 or self.optical_constants_refraction_index == 1)
        self.refractor_image_box_2.setVisible(self.optical_constants_refraction_index == 2 or self.optical_constants_refraction_index == 3)

    def select_prerefl_object_file_name(self):
        self.le_file_prerefl_for_object_medium.setText(oasysgui.selectFileFromDialog(self, self.file_prerefl_for_object_medium, "Select File Prerefl for Object Medium"))

    def select_prerefl_image_file_name(self):
        self.le_file_prerefl_for_image_medium.setText(oasysgui.selectFileFromDialog(self, self.file_prerefl_for_image_medium, "Select File Prerefl for Image Medium"))

    def populate_tab_surface_shape(self, subtab_surface_shape):
        box_interface = oasysgui.widgetBox(subtab_surface_shape, "Surface Shape Parameters", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(box_interface, self, "conic_coefficient_0", "c[1]",  labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_interface, self, "conic_coefficient_1", "c[2]",  labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_interface, self, "conic_coefficient_2", "c[3]",  labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_interface, self, "conic_coefficient_3", "c[4]",  labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_interface, self, "conic_coefficient_4", "c[5]",  labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_interface, self, "conic_coefficient_5", "c[6]",  labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_interface, self, "conic_coefficient_6", "c[7]",  labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_interface, self, "conic_coefficient_7", "c[8]",  labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_interface, self, "conic_coefficient_8", "c[9]",  labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_interface, self, "conic_coefficient_9", "c[10]", labelWidth=260, valueType=float, orientation="horizontal")

        view_shape_box = oasysgui.widgetBox(subtab_surface_shape, "Calculated Surface Shape", addSpace=False, orientation="vertical")

        gui.button(view_shape_box, self, "Render Surface Shape", callback=self.view_surface_shape_data)

    def get_optical_element_instance(self):
        try:     name = self.getNode().title
        except:  name = "Refractive Interface"

        return S4ConicInterface(name=name,
                                boundary_shape=self.get_boundary_shape(),
                                material_object="",  # TODO: discuss with Manolo
                                material_image="", # TODO: discuss with Manolo
                                f_r_ind=self.optical_constants_refraction_index,
                                r_ind_obj=self.refraction_index_in_object_medium,
                                r_attenuation_obj=self.attenuation_in_object_medium,
                                file_r_ind_obj=self.file_prerefl_for_object_medium,
                                r_ind_ima=self.refractive_index_in_image_medium,
                                r_attenuation_ima=self.attenuation_in_image_medium,
                                file_r_ind_ima=self.file_prerefl_for_image_medium,
                                conic_coefficients=[self.conic_coefficient_0,
                                                    self.conic_coefficient_1,
                                                    self.conic_coefficient_2,
                                                    self.conic_coefficient_3,
                                                    self.conic_coefficient_4,
                                                    self.conic_coefficient_5,
                                                    self.conic_coefficient_6,
                                                    self.conic_coefficient_7,
                                                    self.conic_coefficient_8,
                                                    self.conic_coefficient_9])

    def get_beamline_element_instance(self): return S4ConicInterfaceElement()
